// File: ConfigParser.h
// Purpose: Definition of the ConfigParser class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 9/14/2007 12:11:56 PM

#ifndef IDL_CONFIGPARSER_H
#define IDL_CONFIGPARSER_H

#include <string>
#include <sstream>
#include <map>
#include <vector>

namespace idl
{
	/** Class for parsing user configurations. */
	class ConfigParser
	{
	public:
		ConfigParser() {} /**< Constructor. */
		~ConfigParser();  /**< Destructor. */

		void clear(); /**< Clear the current configurations. */

		/**
		 * Print the command line usage.
		 * @param[in]	progName	name of the program
		 */
		void printUsage(const char *progName) const;

		/**
		 * Parse the command line.
		 * @param[in]	argc	number of command line arguments
		 * @param[in]	argv	array of strings for command line arguments
		 * @return		true if success; false if failure
		 */
		bool parseCommandLine(int argc, char *argv[]);

		/**
		 * Parse the config file.
		 * @param[in]	file	name of the config file
		 * @return		true if success; false if failure
		 */
		bool parseConfigFile(const char *file);

		//@{
		/**
		 * Get a string value.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved string value
		 * @return		true if success; false if failure
		 */
		bool getString(int secIdx, const std::string &key, std::string &value) const;
		bool getString(int secIdx, const char *key, std::string &value) const;
		//@}

		//@{
		/**
		 * Get an integer value.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved integer value
		 * @return		true if success; false if failure
		 */
		bool getInt(int secIdx, const std::string &key, int &value) const;
		bool getInt(int secIdx, const char *key, int &value) const;
		//@}

		//@{
		/**
		 * Get an unsigned integer value.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved unsigned integer value
		 * @return		true if success; false if failure
		 */
		bool getUInt(int secIdx, const std::string &key, unsigned int &value) const;
		bool getUInt(int secIdx, const char *key, unsigned int &value) const;
		//@}

		//@{
		/**
		 * Get a real value.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved real value
		 * @return		true if success; false if failure
		 */
		bool getReal(int secIdx, const std::string &key, double &value) const;
		bool getReal(int secIdx, const char *key, double &value) const;
		//@}

		//@{
		/**
		 * Get an integer vector.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	vec		retrieved integer vector
		 * @return		true if success; false if failure
		 */
		bool getIntVec(int secIdx, const std::string &key, std::vector<int> &vec) const;
		bool getIntVec(int secIdx, const char *key, std::vector<int> &vec) const;
		//@}

		//@{
		/**
		 * Get an unsigned integer vector.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	vec		retrieved unsigned integer vector
		 * @return		true if success; false if failure
		 */
		bool getUIntVec(int secIdx, const std::string &key, std::vector<unsigned int> &vec) const;
		bool getUIntVec(int secIdx, const char *key, std::vector<unsigned int> &vec) const;
		//@}

		//@{
		/**
		 * Get a real vector.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	vec		retrieved real vector
		 * @return		true if success; false if failure
		 */
		bool getRealVec(int secIdx, const std::string &key, std::vector<double> &vec) const;
		bool getRealVec(int secIdx, const char *key, std::vector<double> &vec) const;
		//@}

		/**
		 * Get the number of the classifiers to test.
		 * @return number of the classifiers to test
		 */
		unsigned int getClassifierNum() const;

	private:
		/** Type definition for maps storing configurations. */
		typedef std::map<std::string, std::string> ConfigMap;

		/**
		 * Remove comments.
		 * Remove comments and trailing space characters.
		 * @param[in,out]	str	string of a line
		 */
		static void removeComments(std::string &str);

		/**
		 * Split a line into key and value.
		 * @param[in]	line	string of a line
		 * @param[out]	key		retrieved key string
		 * @param[out]	value	retrieved value string
		 */
		static void splitKeyValue(const std::string &line, std::string &key, std::string &value);

		/**
		 * Split a line into fields.
		 * Split a line into fields separated by a specific delimiter.
		 * @param[in]	line		string of a line
		 * @param[out]	fields		retrieved field strings
		 * @param[in]	delimiter	delimieter character (comma by default)
		 */
		static void splitFields(const std::string &line, std::vector<std::string> &fields,
								char delimiter = ',');

		/**
		 * Get a string value from a config map.
		 * @param[in]	map		config map
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved string value
		 * @return		true if success; false if failure
		 */
		static bool getString(const ConfigMap &map, const std::string &key, std::string &value);

		/**
		 * Convert a string to a specific type.
		 * @param[in]	str	string to convert
		 * @param[out]	t	value converted from the given string
		 * @return		true if success; false if failure
		 */
		template <class T>
		static bool strToVal(const std::string &str, T &t)
		{
			std::istringstream iss(str);
			return !(iss >> t).fail();
		}

		/**
		 * Convert a string to a numerical vector of a specific type.
		 * @param[in]	str	string to convert
		 * @param[out]	vec	numerical vector converted from the given string
		 * @return		true if success; false if failure
		 */
		template <class T>
		static bool strToVec(const std::string &str, std::vector<T> &vec)
		{
			vec.clear();

			std::vector<std::string> fields;
			splitFields(str, fields); // Split the string based on ','

			for (std::vector<std::string>::const_iterator itr = fields.begin();
				 itr != fields.end(); ++itr)
			{
				std::vector<std::string> numbers;
				splitFields(*itr, numbers, ':'); // Split the string based on ':'

				if (numbers.size() < 1 || numbers.size() > 3) return false;

				T t1, t2, t3;

				if (!strToVal<T>(numbers[0], t1)) return false;

				if (1 == numbers.size())
				{
					vec.push_back(t1);
				}
				else
				{
					if (2 == numbers.size())
					{
						if (!strToVal<T>(numbers[1], t3)) return false;
						t2 = (t1 <= t3) ? 1 : -1;
					}
					else
					{
						if (!strToVal<T>(numbers[1], t2)) return false;
						if (!strToVal<T>(numbers[2], t3)) return false;
					}

					if (t1 <= t3)
					{
						if (t2 <= 0) return false;

						for (T t = t1; t <= t3; t += t2)
						{
							vec.push_back(t);
						}
					}
					else
					{
						if (t2 >= 0) return false;

						for (T t = t1; t >= t3; t += t2)
						{
							vec.push_back(t);
						}
					}
				}
			}

			return (vec.size() > 0);
		}

		/**
		 * Get a value of a specific type.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	value	retrieved value
		 * @return		true if success; false if failure
		 */
		template <class T>
		bool getT(int secIdx, const std::string &key, T &value) const
		{
			std::string str;
			if (!getString(secIdx, key, str)) return false;
			return strToVal<T>(str, value);
		}

		/**
		 * Get a numerical vector of a specific type.
		 * @param[in]	secIdx	section index (-1 for test configurations)
		 * @param[in]	key		key string
		 * @param[out]	vec		retrieved numerical vector
		 * @return		true if success; false if failure
		 */
		template <class T>
		bool getTVec(int secIdx, const std::string &key, std::vector<T> &vec) const
		{
			std::string str;
			if (!getString(secIdx, key, str)) return false;
			return strToVec<T>(str, vec);
		}
		
	private:
		ConfigMap testConfig; /**< Map storing test configurations. */

		/** Vector of pointers to the maps storing classifier configurations. */
		std::vector<ConfigMap *> classifierConfig;
	};
}

#endif // IDL_CONFIGPARSER_H
