// File: DSDA.cpp
// Purpose: Implementation of the DSDA class
// Remarks: This class is designed for DSDA classification.
// Author: Peter Sadowski
// Modified: October, 2009
 
#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "UIntMatrix.h"
#include "DSDA.h"
#include "LocalSDAModel.h"
#include "math.h"
#include "PolyaFit.h"	
#define INF 1.7E308;
#define REGDEFAULT 1e-3;

using namespace std;

namespace idl
{
	bool DSDA::configure(const SimData &data)
	{
		if (parameters.neighborhoodSize == 0)
			return false;

		// Initialize class variables.
		data.getSimDomain(simDomain);				// Unique sim values over all training data.
		if (simDomain.getRowNum() > 1) {
			simDomain = simDomain.transpose(); }
			
		data.getUniqueLabels(uniqueLabels);			// Unique labels over all training data.

		// Map similarities to their indices in simDomain.
		unsigned int nSimVals = simDomain.size()[1];
		for(unsigned int i = 0; i < nSimVals; ++i) 
		{
			simIdx[simDomain(i)] = i;	// Class variable.
		}
		
		if(parameters.PolyaFlag || parameters.PolyaSmoothedFlag || parameters.AllPairwiseFlag) {
			// Compute Dirichlet parameters by assuming our sim values come from a  Polya distribution.
			if (parameters.AllPairwiseFlag) {
				// Use all local same-class pairwise similarities, rather than those to the class centroid.
				if (!computeAllPairwiseHistograms(data)) {
					cerr << "Error: Failed to calculate the combined histogram." << std::endl;
					return false;
				}
			} else {
				// Compute local pmfs and sum them. Saves both combinedHistogram and localHistMatrices.
		                if (!computeHistograms(data)) {
                		        cerr << "Error: Failed to calculate the combined histogram." << std::endl;
                        		return false;
                		}
			}

			for (vector<int>::const_iterator itrG = uniqueLabels.begin();
				itrG != uniqueLabels.end(); ++itrG)
			{
				for (vector<int>::const_iterator itrH = uniqueLabels.begin();
					itrH != uniqueLabels.end(); ++itrH)
				{
					RealMatrix a;	// Parameter vector.
					vector<int> labelKeyGH(2);
					labelKeyGH[0] = *itrG;
					labelKeyGH[1] = *itrH;
					//OPTION
					if (parameters.AllPairwiseFlag) {
						// Use local histograms of all pairwise similarities.
						polyaFit(localAllPairwiseHistMatrices[labelKeyGH], a);
					} else {
						// Use local histograms of sims to centroids.
						polyaFit(localHistMatrices[labelKeyGH], a);
					}
					// Save these parameters.
					double sumAlphas = a.sum();
					if (sumAlphas != 0) {
						dirichletAlphas[labelKeyGH] = a / sumAlphas;
						// Note that if we want to use smoothing, we do it later (after we incorporate local data).
					} else {
						// We don't have any information.  We'll give alpha's equal weight.
						dirichletAlphas[labelKeyGH] = (a + 1.0/a.length());
					}
					dirichletM[labelKeyGH] = sumAlphas;
					cout << "sum a = " << sumAlphas << std::endl;
				}
			}
		} else if (parameters.locallyRawFlag || parameters.locallySmoothFlag) {
			// Get combined pdfs and exponential pdfs from the combined histogram.
			// Saves results as combinedPMF and combinedDiscExpPMF.
			// Compute local pmfs and sum them. Saves both combinedHistogram and localHistMatrices.
			if (!computeHistograms(data)) {
				cerr << "Error: Failed to calculate the combined histogram." << std::endl;
				return false;
			}
			if (!computeCombinedPDFs()) {
				cerr << "Error: Failed to calculate the combined pmfs." << std::endl;
				return false;
			}
			// Estimate Dirichlet parameters from local and combined histograms.
			if (!computeDirichletParameters(data)) {
				cerr << "Error: Failed to calculate the Dirichlet model parameters." << std::endl;
				return false;
			}
		} else if (parameters.AllPairwiseLSDAFlag) {
			// This method is purely local, so we don't do anything.
		} else {
			cerr << "Error: Unknown Method in DSDA.cpp" << std::endl;
			return false;
		}
		return true;
	}

	//// Calculate (laplace(?) smoothed) class label priors.
	//bool DSDA::computeSmoothedPriors(const SimData &data)
	//{
	//	cerr << "Error: this method is deprecated." << std::endl;
	//	assert(false);
	//	vector<int> trainingLabels;
	//	data.getTrainingLabels(trainingLabels);		
	//	unsigned int nTraining = data.getTrainingSetSize();
	//	map<int, double> smoothedLabelPriors;
	//	for (vector<int>::const_iterator itrG = uniqueLabels.begin();
	//		itrG != uniqueLabels.end(); ++itrG)
	//	{
	//		// PJS: This causes a warning that we can ignore with reasonable input. 
	//		// google "because difference_type is guaranteed to be large enough to hold the answer."
	//		//int Gcount = count(trainingLabels.begin(), trainingLabels.end(), *itrG);
	//		//smoothedLabelPriors[*itrG] = (Gcount + 1.0)/(nTraining + uniqueLabels.size());
	//	}
	//	return true;
	//}

	// Compute combined histogram of G-HCentroid similarity values over all local neighborhoods.
	// Also computes localPMFs, a pmf for each class in each local neighborhood.
	bool DSDA::computeHistograms(const SimData &data)
	{
		// Training labels.	
		vector<int> trainingLabels;
		data.getTrainingLabels(trainingLabels);
		unsigned int nTraining = data.getTrainingSetSize();
		unsigned int nSimVals = simDomain.size()[1];
		map<int, unsigned int> labelCounts;
		computeVectorCounts(trainingLabels, labelCounts);

		// First initialize histograms to zero.
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
			itrG != uniqueLabels.end(); ++itrG)
		{
			for (vector<int>::const_iterator itrH = uniqueLabels.begin();
				itrH != uniqueLabels.end(); ++itrH)
			{
				// Key for identifying current G,H pair.
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;
				UIntMatrix temp(1,nSimVals, 0);
				combinedHistogram[labelKeyGH] = temp;
				// Matrix where each row is a local histogram.
				UIntMatrix temp2(nTraining, nSimVals, 0);
				localHistMatrices[labelKeyGH] = temp2;
			}
		}

		// Iterate over each training samples, adding items to histograms.
		for (unsigned int i = 0; i < nTraining; ++i) 
		{
			// Get local sim matrix.
			RealMatrix simMat(parameters.neighborhoodSize, parameters.neighborhoodSize);
			data.getNNofTrainingSimMat(i, parameters.neighborhoodSize, simMat);
			// Get local sim vector.
			RealMatrix simVec(1, parameters.neighborhoodSize);
			data.getNNofTrainingSimVec(i, parameters.neighborhoodSize, simVec);
			// Neighborhood labels.
			vector<int> labels(parameters.neighborhoodSize);
			data.getNNofTrainingLabels(i, parameters.neighborhoodSize, labels);
			// Construct local SDA model for this training sample.
			LocalSDAModel lSDA;
			lSDA.LocalSDAModel::initialize(simMat, labels, simVec, simDomain);

			for (vector<int>::const_iterator itrG = uniqueLabels.begin(); itrG != uniqueLabels.end(); ++itrG)
			{

				for (vector<int>::const_iterator itrH = uniqueLabels.begin(); itrH != uniqueLabels.end(); ++itrH)
				{
					// Get the local similarities between local samples in class G and H centroid.  (Doesn't include training sample i)
					RealMatrix localHCenterToGSims;
					lSDA.getHCenterToGSims(simMat,localHCenterToGSims,*itrH, *itrG);
					if (!localHCenterToGSims.isEmpty())
					{
						// Get local histogram.
						UIntMatrix localHistogram(1,nSimVals, 0);
						for (unsigned int j = 0; j < localHCenterToGSims.length(); ++j)
						{
							localHistogram(simIdx[localHCenterToGSims(j)]) += 1;
						}

						// Key for histograms (maps from G,H pair).
						vector<int> labelKeyGH(2);
						labelKeyGH[0] = *itrG;
						labelKeyGH[1] = *itrH;
						// Key for individual pmfs (maps from sample,H pair).
						vector<int> pmfKey(3);
						pmfKey[0] = i;					// Index of the training sample, neighborhood.
						pmfKey[1] = *itrG;				// Class label 1.
						pmfKey[2] = *itrH;				// Class label 2.
						// Temp holder for real pmfs.
						RealMatrix tempLocalHistReal;
						tempLocalHistReal.zeros(1,nSimVals);
						for (unsigned int j = 0; j < nSimVals; ++j) 
						{
							// Update combined histogram.
							combinedHistogram[labelKeyGH](j) += localHistogram(j);	// PJS: Fixed from '=' on Nov30th.
							// Update local histogram matrix.
							localHistMatrices[labelKeyGH](i, j) = localHistogram(j);	// Slow because reallocating memory.
							// For updating individual pmfs below.
							tempLocalHistReal(j) = localHistogram(j);
						}

						// Save individual pmfs.
						if (tempLocalHistReal.sum() != 0) {
							localPMFs[pmfKey] = tempLocalHistReal / tempLocalHistReal.sum();
							// Save individual discrete exponential smoothed pmfs.
							localDiscExpPMFs[pmfKey] = idl::getDiscreteExponential(localPMFs[pmfKey], simDomain);
						} // Else: don't do anything.  We don't care if we don't have data, because this pmf will be ignored.		
					}				
				}  // End iteration over H.
			}	// End iteration over G.
		}	// End iteration over samples.
		return true;
	}
	
	// This function computes localAllPairwiseHistograms
	// for the "AllPairwise" method, in which we assume all g-h similarities come from the gh Polya distribution.	
	bool DSDA::computeAllPairwiseHistograms(const SimData &data) 
	{	
		// Training labels and sims.	
		vector<int> trainingLabels;
		data.getTrainingLabels(trainingLabels);
		unsigned int nTraining = data.getTrainingSetSize();
		unsigned int nSimVals = simDomain.size()[1];
		map<int, unsigned int> labelCounts;
		computeVectorCounts(trainingLabels, labelCounts);
		
		// First initialize pairwise count histogram matrices to zero.
		for (vector<int>::const_iterator itrG = uniqueLabels.begin(); itrG != uniqueLabels.end(); ++itrG) 
		{
			for (vector<int>::const_iterator itrH = uniqueLabels.begin(); itrH != uniqueLabels.end(); ++itrH) 
			{
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;
				// Matrix where each row is a local histogram.
				UIntMatrix temp2(nTraining, nSimVals, 0);
				localAllPairwiseHistMatrices[labelKeyGH] = temp2;
			}
		}

		// Iterate over each training sample, adding items to histograms.
		for (unsigned int i = 0; i < nTraining; ++i) 
		{
			// First initialize a lSDA model.  This calculates the samplesByLabel mapping for us.
			// Get local sim matrix.
			RealMatrix simMat(parameters.neighborhoodSize, parameters.neighborhoodSize);
			data.getNNofTrainingSimMat(i, parameters.neighborhoodSize, simMat);
			// Get local sim vector.
			RealMatrix simVec(1, parameters.neighborhoodSize);
			data.getNNofTrainingSimVec(i, parameters.neighborhoodSize, simVec);
			// Neighborhood labels.
			vector<int> labels(parameters.neighborhoodSize);
			data.getNNofTrainingLabels(i, parameters.neighborhoodSize, labels);
			// Construct local SDA model for this training sample.
			LocalSDAModel lSDA;
			lSDA.LocalSDAModel::initialize(simMat, labels, simVec, simDomain);	// Calculates the samplesByLabel mapping for us. 

			for (vector<int>::const_iterator itrG = uniqueLabels.begin(); itrG != uniqueLabels.end(); ++itrG)
			{
				for (vector<int>::const_iterator itrH = uniqueLabels.begin(); itrH != uniqueLabels.end(); ++itrH)
				{
					// NOTE: localAllPairwiseHistMatrices(G,H) = localAllPairwiseHistMatrices(H,G) since
					//       sims are assumed to be symmetrical (I.E. we calculate these twice)
					//       But I'll just let it go for now.   -Peter 
					vector<int> labelKeyGH(2);
					labelKeyGH[0] = *itrG;
					labelKeyGH[1] = *itrH;

					// Get the local similarities between samples in g and h.  
					RealMatrix ghSimMatrix;  // Matrix with each row a class-g sample and each column a class-h sample (all local).
					vector<unsigned int> vecGIndices;
					vector<unsigned int> vecHIndices;
					lSDA.getSamplesByLabel(vecGIndices, *itrG);
					lSDA.getSamplesByLabel(vecHIndices, *itrH);
					ghSimMatrix = simMat(vecGIndices, vecHIndices);
					vector<double> ghSimVector;
					ghSimMatrix.dump(ghSimVector);  // Vector of all observed g-to-h similarity values for this neighborhood.
					if (*itrG == trainingLabels[i]) {
						// Add sims of the training sample i if it's from class g.  Simply push these on to the ghSimVector.
						vector<double> temp;
						simVec(0,vecHIndices).dump(temp);
						ghSimVector.insert(ghSimVector.end(), temp.begin(), temp.end());
					} else if (*itrH == trainingLabels[i]) {
						// Add sims of training sample i if it's from class h. (Assumes similarity is symmetrical.)
						vector<double> temp;
						simVec(0, vecGIndices).dump(temp);
						ghSimVector.insert(ghSimVector.end(), temp.begin(), temp.end());
					}

					//  Add observed G-H similarities in this neighborhood to our data structure.
					for (unsigned int k = 0; k < ghSimVector.size(); ++k)
					{
						// Matrix indexed by (g,h) pair.
						// Index within matrix by neighborhood (row) and simvalue (column).
						localAllPairwiseHistMatrices[labelKeyGH](i, simIdx[ghSimVector.at(k)]) += 1;
					}		
				} // End iteration over h.
			} // End iteration over g.
		} // End iteration over samplse. 	
		return true;
	}



	// Calculate the combined pmfs, and the combined discrete exponential pmfs.
	// Does this using the combinedHistogram data.
	bool DSDA::computeCombinedPDFs()
	{
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
			itrG != uniqueLabels.end(); ++itrG)
		{
			for (vector<int>::const_iterator itrH = uniqueLabels.begin();
				itrH != uniqueLabels.end(); ++itrH)
			{
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;

				assert(combinedHistogram[labelKeyGH].getRowNum() == 1);
				combinedPMF[labelKeyGH].resize(combinedHistogram[labelKeyGH].getRowNum(),
											   combinedHistogram[labelKeyGH].getColumnNum());
				// Calculate a pmf from the combined histogram.
				unsigned int combinedNumSamples = 0;
				for (unsigned int i = 0; i < combinedHistogram[labelKeyGH].length(); ++i) {
					combinedPMF[labelKeyGH](i) = combinedHistogram[labelKeyGH](i);
					combinedNumSamples += combinedHistogram[labelKeyGH](i);
				}
				if (combinedNumSamples != 0) {
					combinedPMF[labelKeyGH] /= combinedNumSamples;
					// Calculate the combined discrete exponential pmf.
					combinedDiscExpPMF[labelKeyGH] = idl::getDiscreteExponential(combinedPMF[labelKeyGH], simDomain);
				} else {
					// We don't have any data.  Assume uniform distribution.
					combinedPMF[labelKeyGH].zeros(1, combinedPMF[labelKeyGH].length());
					combinedPMF[labelKeyGH] += 1.0/combinedPMF[labelKeyGH].length();
					// It doesn't seem right to use anything but a flat curve if we don't have data.
					combinedDiscExpPMF[labelKeyGH] = combinedPMF[labelKeyGH];
				}
			}
		}
		return true;
	}

	// Compute the Dirichlet parameters from a bunch of estimated pmfs,
	// assumed to come from the Dirichlet, using method of moments.  
	// The mean is just the average pmf; the tentpole parameter is the harder part.
	bool DSDA::computeDirichletParameters(const SimData &data) 
	{
		// Epsilon is used for numerical issues.
		double epsilon = 1E-8;	// Can be pretty important, especially with 1NN.
		unsigned int nSimVals = simDomain.size()[1];
		vector<int> trainingLabels;
		data.getTrainingLabels(trainingLabels);
		unsigned int nTraining = data.getTrainingSetSize();
		// Estimate a Dirichlet prior for each label pair.
		for (vector<int>::const_iterator itrG = uniqueLabels.begin();
			itrG != uniqueLabels.end(); ++itrG)
		{
			for (vector<int>::const_iterator itrH = uniqueLabels.begin();
				itrH != uniqueLabels.end(); ++itrH)
			{
				vector<int> labelKeyGH(2);
				labelKeyGH[0] = *itrG;
				labelKeyGH[1] = *itrH;
				
				// Choose local pmfs to use when estimating the variance. (localPMFs or localDiscExpPMFs)
				map<vector<int>, RealMatrix> chosenLocalPMF;
				if (parameters.locallyRawFlag) {
					assert(parameters.locallySmoothFlag == false);
					chosenLocalPMF = localPMFs;
				}
				else if(parameters.locallySmoothFlag) {
					chosenLocalPMF = localDiscExpPMFs;
				} else {
					cerr << "No local pmf flag set." << endl;
					return false;
				}
				
				// Calculate E[p_k] and E[p_k^2] vectors from training data local pmfs.
				RealMatrix meanPKVec;
				RealMatrix meanPKSquaredVec;
				meanPKVec.zeros(1,nSimVals);		// Initialize as zero because we are summing.
				meanPKSquaredVec.zeros(1,nSimVals);		// Initialize as zero because we are summing.				
				vector<int> pmfKey(3);
				pmfKey[1] = *itrG; // pmfKey[0] specified below.
				pmfKey[2] = *itrH;
				unsigned int pmfCount = 0; 
				for (unsigned int i = 0; i < nTraining; ++i) {
					if (trainingLabels[i] != *itrG)
						continue;
					pmfKey[0] = i;
					if (chosenLocalPMF.count(pmfKey) == 1) {
						// The ith neighborhood has a centroid H and samples from G, so we have a pmf.
						RealMatrix pmf = chosenLocalPMF[pmfKey];
						assert(meanPKSquaredVec.getRowNum() == pmf.getRowNum());
						meanPKVec += pmf;					// p_k's are summed.
						meanPKSquaredVec += pmf.dot(pmf);	// p_k's are squared and summed.
						pmfCount++;
					} else {
						// There are no representatives from G and H in the neighborhood of our training point i.
						// (This doesn't include the training point itself.)
						assert(chosenLocalPMF.count(pmfKey) == 0);
					}
				}

				double alphaSum;
				if (pmfCount == 0) {
					// We have no data to work with.
					alphaSum = 0;
				} else {
					// Estimate the Dirichlet tentpole parameter based on variance of PK's.
					meanPKVec /= pmfCount;
					meanPKSquaredVec /= pmfCount; 
					unsigned int bigKminus1 = meanPKVec.length()-1;		// K-1 in Minka2003 eq 23.
					// Calculate log sum alphas using Minka2003 Eq 23.
					// This equivalent form is easier to deal with.
					alphaSum = 1;
					for (unsigned int k = 0; k < bigKminus1; ++k) { 
						double E_PK = meanPKVec(k);				// E[p_k]
						double E_PK2 = meanPKSquaredVec(k);		// E[p_k^2]
						double E_PK_2 = pow(E_PK,2);			// E[p_k]^2
						if (E_PK < epsilon && E_PK2 < epsilon) {
							// We can ignore this k.
							continue;
						} else if (!((E_PK + epsilon >= E_PK2) && (E_PK2 + epsilon >= E_PK_2))) {	// Should be harsher?
							cerr << "Error: shouldn't happen." << std::endl;
							return false;
						}
						double temp = (E_PK - E_PK2 + epsilon) / (E_PK2 - E_PK_2 + epsilon);
						temp = pow(temp, (1.0/bigKminus1));
						alphaSum *= temp;
					}
				}
				
				// SET DIRICHLET PARAMETERS
				double c = alphaSum;	// Tentpole parameter.
				// c = 0; would reduce this to SDA.
				dirichletM[labelKeyGH] = c;
				dirichletAlphas[labelKeyGH] = meanPKVec;		//Dec11th Unweighted, simple average of (non-zero) local pmfs. (raw or smoothed)
				if (parameters.locallyRawFlag) {
					std::cout << "raw c " << c << std::endl;
				}
				else if(parameters.locallySmoothFlag) {
					std::cout << "smooth c " << c << std::endl;
				} else {
					cerr << "No local pmf flag set." << endl;
					return false;
				}
				
				// Nov30th, used the smoothed "weighted" mean of histograms.
				//dirichletAlphas[labelKeyGH] = combinedDiscExpPMF[labelKeyGH];//combinedPMF or combinedDiscExpPMF;		
				
			}
		}
		return true;
	}


	// Classify the test samples of the data set, and get their class probabilities.
	bool DSDA::classify(const SimData &data, vector<int> &results, RealMatrix &probMatrix)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size.
		if (data.getTrainingSetSize() < parameters.neighborhoodSize) {
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}
		// Initialize result data structures.
		unsigned int numTestSamples = data.getTestSetSize();
		results.resize(numTestSamples);
		unsigned int nUniqueLabels = (unsigned int)uniqueLabels.size();
		probMatrix.resize(numTestSamples,nUniqueLabels);	// Class probabilities for each test point.

		// For each test sample, classify and get class probabilities.
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			// Initialize neighborhood-specific data structures.
			RealMatrix simMat(parameters.neighborhoodSize,
						parameters.neighborhoodSize);				// Local sim matrix of neighborhood.
			vector<int> labels(parameters.neighborhoodSize);		// Labels of neighborhood.
			RealMatrix simVec(1, parameters.neighborhoodSize);		// Sim of test point to neighborhood.
			data.getNNSimMat(i, parameters.neighborhoodSize, simMat);
			data.getNNLabels(i, parameters.neighborhoodSize, labels);
			data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
			// Use the local SDA model for some convenient data structures.
			LocalSDAModel unregLsdaModel;
			unregLsdaModel.initialize(simMat, labels, simVec, simDomain);
			if(unregLsdaModel.haveReasonToUseBackupClassifier()) {
				// Use a backup classifier.
				map<int, double> tempProbs;
				results[i] = unregLsdaModel.classify(simMat, labels, simVec, simDomain, tempProbs);
				for (unsigned int labelIdx = 0; labelIdx < uniqueLabels.size(); ++labelIdx) {
					if (tempProbs.count(uniqueLabels[labelIdx])) {
						probMatrix(i,labelIdx) = tempProbs[uniqueLabels[labelIdx]]; }
				}
				if (probMatrix.sumRow(i) < .9999 || probMatrix.sumRow(i) > 1.0001) {		
					cerr << "Probs not summing to 1 in DSDA classifier." << endl;
					return false; }
				continue;	
			}
			// Compute these only if necessary.
			unregLsdaModel.computeModelParams(simMat, labels, simVec, simDomain);  
			vector<int> uniqueLocalLabels;			// List of unique labels in this neighborhood.
			unregLsdaModel.getUniqueLabels(uniqueLocalLabels);

			// Iterate over neighborhood class labels G.
			double maxLikelihood = (-1)*INF;
			int maxLikelihoodLabel = 0;
			RealMatrix classLikelihoods_i;	// Holds the class probabilities of testpoint i.						
			classLikelihoods_i.zeros(1,nUniqueLabels);
			for (vector<int>::const_iterator itrG = uniqueLocalLabels.begin(); 
				itrG != uniqueLocalLabels.end(); ++itrG)
			{
				// Calculate likelihood of testpoint being in class G.
				double tempLikelihood;
				if(parameters.AllPairwiseFlag || parameters.AllPairwiseLSDAFlag) {
					tempLikelihood = posteriorAllPairwiseLikelihood(data, unregLsdaModel, *itrG, i);
				} else {
					tempLikelihood = posteriorDirichletLikelihood(data, unregLsdaModel, *itrG, i);
				}

				if (tempLikelihood < 0) {
					cerr << "Error: Likelihood < 0." << std::endl;
					return false;
				}
				// Keep track of the most likely label.
				if (tempLikelihood > maxLikelihood) {
					maxLikelihood = tempLikelihood;
					maxLikelihoodLabel = *itrG;
				}
				// Save likelihood of this label G to a vector.
				for (unsigned int j = 0; j < nUniqueLabels; ++j)
				{
					if (uniqueLabels[j] == *itrG)
					{
						// j is the index of label *itrG. 
						classLikelihoods_i(j) = tempLikelihood;
					}
				}
			}  // End iteration over G.
			// Normalize likelihoods to get probabilities.
			classLikelihoods_i /= classLikelihoods_i.sum();
			// Record results and probabilities.
			for (unsigned int j = 0; j < nUniqueLabels; j++)
				probMatrix(i,j) = classLikelihoods_i(j);
			results[i] = maxLikelihoodLabel;
			// Check that probabilities sum to 1.
			if (probMatrix.sumRow(i) < .9999 || probMatrix.sumRow(i) > 1.0001)
			{		
				cerr << "Probs not summing to 1 in DSDA classifier." << endl;
				return false;
			}
		} // End iteration over test points.
		return true;
	}

	// Classify the test samples of the data set.
	bool DSDA::classify(const SimData &data, vector<int> &results)
	{
		RealMatrix probMatrix;
		return classify(data,results,probMatrix);
	}

	/**
	 *
	 * Returns the posterior likelihood of seeing this samples' similarities to each centroid, 
	 * assuming the sample is in class g.
	 * likelihood = prior(g)  * Prod_h( P(s(x, mu_h)| Y=g))
	 */
	double DSDA::posteriorDirichletLikelihood(const SimData &data,
											  LocalSDAModel &unregLsdaModel,
											  const int g, int testSampleIndex)
	{
		// Get necessary data structures.
		RealMatrix simMat(parameters.neighborhoodSize,
			parameters.neighborhoodSize);				// Local sim matrix of neighborhood.
		RealMatrix simVec(1, parameters.neighborhoodSize);		// Sim of test point to neighborhood.
		data.getNNSimMat(testSampleIndex, parameters.neighborhoodSize, simMat);
		data.getNNSimVec(testSampleIndex, parameters.neighborhoodSize, simVec);
		vector<int> uniqueLocalLabels;	// List of unique labels in this neighborhood.
		unregLsdaModel.getUniqueLabels(uniqueLocalLabels);
		map<int , double> simTestToHCentroid;	// Map from local label H to s(testpoint, centroidH);
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH) {
				// Get similarity between test sample and centroid of H.
				simTestToHCentroid[*itrH] = unregLsdaModel.getSimTestToH(simVec, *itrH);
		}
		
		// Iterate over the unique local labels H.
		double likelihood = unregLsdaModel.getPrior(g);  // Starting likelihood.
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			// Calculate the likelihood of seeing the observed similarities to centroid H.
			vector<int> keyGH(2);	// Key for simplifying calls to matrices.
			keyGH[0] = g;
			keyGH[1] = *itrH;

			// Get histogram of local G samples to centroid H.
			RealMatrix simsGToHCentroid; // Vector of similarities from centroid H to local samples of class G.	
			unregLsdaModel.getHCenterToGSims(simMat,simsGToHCentroid,*itrH, g);
			// Convert simsGToHCentroid to histogram form, over the entire domain of sim values.
			RealMatrix histogramsGtoHCenter;	
			histogramsGtoHCenter.zeros(1,simDomain.length());
			for (unsigned int i = 0; i < simsGToHCentroid.length(); ++i) {
				histogramsGtoHCenter(simIdx[simsGToHCentroid(i)]) += 1;
			}
			
			// alpha_k = alpha_0 * E[p_k] in wikipedia Dirichlet notation.
			RealMatrix mAlpha = dirichletM[keyGH] * dirichletAlphas[keyGH]; 
			// Get the expected distribution. (i.e. Dirichlet prior combined with local histogram.)
			double m_tilde = dirichletM[keyGH] + histogramsGtoHCenter.sum();
			assert(m_tilde != 0);
			assert(mAlpha.getColumnNum() == histogramsGtoHCenter.getColumnNum());
			assert(mAlpha.getRowNum() == histogramsGtoHCenter.getRowNum());
			RealMatrix alpha_tilde = mAlpha + histogramsGtoHCenter;
			alpha_tilde /= m_tilde;

			//Stopped smoothing posterior on Dec14th.	
			//if (parameters.PolyaSmoothedFlag || parameters.locallyRawFlag || parameters.locallySmoothFlag) {
			if (parameters.PolyaSmoothedFlag) {
				// Smooth out our posterior distribution.  (Standard Polya does not do this.)
				alpha_tilde = idl::getDiscreteExponential(alpha_tilde, simDomain);
			}
			if (!(alpha_tilde.sum() > .9999999 && alpha_tilde.sum() < 1.00000001)) {
				double temp = alpha_tilde.sum(); 
			}

			// Probability of the observed sim value.
			double prob = alpha_tilde(simIdx[simTestToHCentroid[*itrH]]);
			assert(prob >= 0  && prob <= 1); // Anything named "prob" should be in the range [0,1];
			// Update likelihood.
			likelihood *= prob;
		}
		assert(likelihood >= 0);
		return likelihood;
	}


	/**
	 *
	 * Returns the posterior likelihood of seeing this samples' average similarity to the samples in each class, 
	 * assuming the sample is in class g.
	 * likelihood = prior(g)  * Prod_h( P( mean(s(x, z_h)) | Y=g)) 
	 *	
	 * Depending on parameters, either uses only the local g-h similarity data or the Dirichlet prior. 
	 */
	double DSDA::posteriorAllPairwiseLikelihood(const SimData &data,
											  LocalSDAModel &unregLsdaModel,
											  const int g, int testSampleIndex)
	{
		// Get necessary local data structures.
		RealMatrix simMat(parameters.neighborhoodSize,
			parameters.neighborhoodSize);				// Local sim matrix of neighborhood.
		RealMatrix simVec(1, parameters.neighborhoodSize);		// Sim of test point to neighborhood.
		data.getNNSimMat(testSampleIndex, parameters.neighborhoodSize, simMat);
		data.getNNSimVec(testSampleIndex, parameters.neighborhoodSize, simVec);
		vector<int> uniqueLocalLabels;	// List of unique labels in this neighborhood.
		unregLsdaModel.getUniqueLabels(uniqueLocalLabels);
		
		// Starting likelihood. To be multiplied by class likelihoods.
		double likelihood = unregLsdaModel.getPrior(g);  
		// Likelihood of the test sample's observed similarities to class h samples. (Avg of pairwise sim likelihoods.)
		vector<double> classLikelihoods;

		// Iterate over the unique local labels H.
		for (vector<int>::const_iterator itrH = uniqueLocalLabels.begin(); 
			itrH != uniqueLocalLabels.end(); ++itrH)
		{
			// Calculate the average likelihood of seeing the observed similarities to samples in class h.
			// Then multiply the likelihood by this average likelihood.
			vector<int> keyGH(2);	// Key for simplifying calls to matrices.
			keyGH[0] = g;
			keyGH[1] = *itrH;

			// Get the local similarities between samples in g and h.  
			RealMatrix ghSimMatrix;  // Matrix with each row a class-g sample and each column a class-h sample (all local).
			vector<unsigned int> vecGIndices;
			vector<unsigned int> vecHIndices;
			unregLsdaModel.getSamplesByLabel(vecGIndices, g);
			unregLsdaModel.getSamplesByLabel(vecHIndices, *itrH);
			ghSimMatrix = simMat(vecGIndices, vecHIndices);
			vector<double> ghSimVector;
			ghSimMatrix.dump(ghSimVector);  // Vector of all observed g-to-h similarity values for this neighborhood.
			RealMatrix localHistogram(1,simDomain.length(), 0);
			for (unsigned int j = 0; j < ghSimVector.size(); ++j)
			{
				localHistogram(simIdx[ghSimVector.at(j)]) += 1;
			}
			
			RealMatrix pmf;	// PMF over similarity domain.
			if (parameters.AllPairwiseLSDAFlag)
			{
				// AllPairwiseLSDA only uses the locally observed data (no prior).
				pmf = localHistogram / localHistogram.sum();
				pmf = idl::getDiscreteExponential(pmf, simDomain);
			} else if (parameters.AllPairwiseFlag) {
				// Use the Dirichlet prior calculated upon configuration and the local data to calculate Bayes posterior.
				// alpha_k = alpha_0 * E[p_k] in wikipedia Dirichlet notation.
				RealMatrix mAlpha = dirichletM[keyGH] * dirichletAlphas[keyGH]; 
				// Get the expected distribution. (i.e. Dirichlet prior combined with local histogram.)
				double m_tilde = dirichletM[keyGH] + localHistogram.sum();
				assert(m_tilde != 0);
				assert(mAlpha.getColumnNum() == localHistogram.getColumnNum());
				assert(mAlpha.getRowNum() == localHistogram.getRowNum());
				RealMatrix alpha_tilde = mAlpha + localHistogram;
				alpha_tilde /= m_tilde;	// alpha_tilde is the mean pmf of our posterior Dirichlet.
				if (!(alpha_tilde.sum() > .9999999 && alpha_tilde.sum() < 1.00000001)) {
					double temp = alpha_tilde.sum(); 
					assert(false);
				}
				pmf = alpha_tilde;
			}

			// Get the local similarities between training sample and local h samples.  
			RealMatrix hSimVector;  // Matrix with 1 row (test sample) and each column a class-h sample (all local).
			hSimVector = simVec(0, vecHIndices);

			// Calculate AVERAGE likelihood of the observed sim values to class h samples.
			double avgProb = 0;
			for (unsigned int j = 0; j < hSimVector.length() ; ++j)
			{
				avgProb +=  pmf(simIdx[hSimVector(j)]);
			} 
			avgProb /=  hSimVector.length();
			assert(avgProb >= 0  && avgProb <= 1);	// Anything named "prob" should be in the range [0,1];
			classLikelihoods.push_back(avgProb);
			 
			// Update total likelihood.
			likelihood *= avgProb;
		}
		assert(likelihood >= 0);
		return likelihood;
	}


	// Count the number of times each integer appears in the vector.
	void DSDA::computeVectorCounts(const vector<int> &labels, map<int, unsigned int> &counts)
	{
		for(vector<int>::const_iterator itr = labels.begin();
			itr != labels.end(); ++itr)
		{
			if(counts.find(*itr)==counts.end()) {
				// Haven't seen this label yet.
				counts[*itr] = 1;
			} else {
				// Have seen this label before.
				counts[*itr]++;
			}
		}
	}
}
