// File: DSDA.h
// Purpose: Implementation of the DSDA class
// Remarks: This class is designed for DSDA classification.
// Author: Peter Sadowski
// Modified: October, 2009

#ifndef IDL_DSDA_H
#define IDL_DSDA_H

#include <vector>
#include <map>
#include <fstream>
#include "RealMatrix.h"
#include "UIntMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "LocalSDAModel.h"
#include <string>
namespace idl
{
	/**
	 * Class for the Local SDA classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class DSDA : public Classifier
	{
	public:
		/** Structure type for some parameters. */
		struct ParaType
		{
			unsigned int neighborhoodSize;	/**< Neighborhood size. */
			bool locallyRawFlag;
			bool locallySmoothFlag;
			bool PolyaFlag;
			bool PolyaSmoothedFlag;
			bool AllPairwiseFlag;
			bool AllPairwiseLSDAFlag;
		};

		/** Default constructor. */
		DSDA()
		{
			parameters.neighborhoodSize = 1;
			parameters.locallyRawFlag = false;	
			parameters.locallySmoothFlag = false;
			parameters.PolyaFlag = false;
			parameters.PolyaSmoothedFlag = false;
			parameters.AllPairwiseFlag = false;
			parameters.AllPairwiseLSDAFlag = false;
		}

		/**
		 * Constructor.
		 * @param	para	Parameters for the classifier.
		 */
		DSDA(const ParaType &para)
		{
			parameters = para;
		}

		/** Destructor. */
		~DSDA() {}	

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSize(unsigned int k)
		{
			if (k > 0)
			{
				parameters.neighborhoodSize = k;
			}
		}

		/**
		 * These flags are determined by the mode in the config file.
		 */
		void setFlag(bool locallyRawFlag, bool locallySmoothFlag, bool PolyaFlag, bool PolyaSmoothedFlag, bool AllPairwiseFlag, bool AllPairwiseLSDAFlag)
		{
			parameters.locallyRawFlag = locallyRawFlag;
			parameters.locallySmoothFlag = locallySmoothFlag;
			parameters.PolyaFlag = PolyaFlag;
			parameters.PolyaSmoothedFlag = PolyaSmoothedFlag;
			parameters.AllPairwiseFlag = AllPairwiseFlag;
			parameters.AllPairwiseLSDAFlag = AllPairwiseLSDAFlag;
		}

		/**
		 * Set the parameters for the classifier.
		 * @param[in]	para	parameters for the Local SDA classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}
		
		// Configure the classifier (Use the data to estimate a dirichlet.).
		bool configure(const SimData &data);

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		Similarity data set.
		 * @param[out]	results		Vector to store the estimated class labels.
		 * @param[out]  probMatrix  nTestSamples*nUniqueLabels matrix of probabilities.
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);
		bool classify(const SimData &data, std::vector<int> &results, RealMatrix &probMatrix);

	private:
		/**
		 * Check if the given labels belong to one class.
		 * @param[in]	labels		vector of the given class labels
		 * @return		true if the given labels belong to one class;
		 *				false if they belong to more than one class
		 */
		static bool isOneClass(const std::vector<int> &labels);
		
		// Computes posterior likelihood from Dirichlet model.
		// Assumes Dirichlet parameters were estimated based on sample-to-centroid sims.
		double posteriorDirichletLikelihood(const SimData &data,
											LocalSDAModel &unregLsdaModel,
											const int g, int testSampleIndex);
		// Computes a likelihood using all pairwise g-to-h sims.
		// Depending on parameters, this may be the Dirichlet posterior, or just the local, smoothed version.
		double posteriorAllPairwiseLikelihood(const SimData &data,
											  LocalSDAModel &unregLsdaModel,
											  const int g, int testSampleIndex);
		//bool computeSmoothedPriors(const SimData &data);
		bool computeHistograms(const SimData &data);
		bool computeAllPairwiseHistograms(const SimData &data);
		bool computeCombinedPDFs();
		bool computeDirichletParameters(const SimData &data);
		RealMatrix getDiscreteExponential(const RealMatrix &rawPMF);
		void computeVectorCounts(const std::vector<int> &labels, std::map<int, unsigned int> &counts);
		

		// Parameters are initialized in constructor.
		ParaType parameters;
		// Private class variables that are initialized upon configuration.
		RealMatrix simDomain;					// Vector of sim values.		
		std::map<double,unsigned int> simIdx;		// Mapping from the sim values to their index in simDomain. 
		std::vector<int> uniqueLabels;				// Vector of the unique labels in the training set. 
		std::map<std::vector<int>, UIntMatrix> combinedHistogram;	// Hist of all G-HCent sim values. 
		std::map<std::vector<int>, RealMatrix> combinedPMF;			// PMF of all G-HCent sim values.
		std::map<std::vector<int>, RealMatrix> combinedDiscExpPMF;	// Discrete smoothed pmf of all G-HCent sim vals.
		// These data structures map a <trainingSampleIndex, class G, class H> triple 
		// to pmfs of sim values from class G samples to HCentroid in the neighborhoods around the training sample.
		std::map<std::vector<int>, RealMatrix> localPMFs;
		std::map<std::vector<int>, RealMatrix> localDiscExpPMFs;
		// This data structure maps <class G, class H> pair to matrix of histograms.
		std::map<std::vector<int>, UIntMatrix> localHistMatrices;
		/* This data structure maps <class G, class H> pair to matrix of histograms.
		 * Each matrix has |simDomain| columns and a row for each training sample of class g.
		 */
		std::map<std::vector<int>, UIntMatrix> localAllPairwiseHistMatrices;
		//Deprecated std::map<int,double> labelPriors;

		// The Dirichlet parameters.
		std::map<std::vector<int>, double> dirichletM;
		std::map<std::vector<int>, RealMatrix> dirichletAlphas;	// PMF

		// Polya parameters.
		std::map<std::vector<int>, RealMatrix> polyaAlphas;

	};
}

#endif // IDL_LocalSDA_H
