// File: DSDATest.cpp
// Purpose: Implementation of the DSDATest class.
// Remarks: This class is designed as an adapter for the DSDA class
//			to enable testing on multiple sets of parameters.
// Author: Peter Sadowski, Ingrid Liu

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "DSDATest.h"

using namespace std;

namespace idl
{
	const string DSDATest::_ID = "DSDA"; // Define the class ID

	// Register the class instantiator with the factory
	const bool DSDATest::_REGISTERED =
		ClassifierTestFactory::registerCreator(DSDATest::_ID, &DSDATest::create);

	// Destructor
	DSDATest::~DSDATest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a weighted k-NN classifier.
	bool DSDATest::configure(ConfigParser &config, int secIdx)
	{
		std::cout << "configuring... \n";
		// Get mode for DSDA
		// locallyRaw, locallySmooth, Polya
		string strMode;
		if (!config.getString(secIdx, "Mode", strMode))
		{
			cerr << "Missing mode for DSDA classifier " << secIdx << endl;
			return false;
		}
		setFlags(strMode, locallyRawFlag, locallySmoothFlag, PolyaFlag, PolyaSmoothedFlag, AllPairwiseFlag, AllPairwiseLSDAFlag);

		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for kNN classifier "
				 << secIdx << endl;
			return false;
		}
		//Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for rlSDAclassifier "
				 << secIdx << endl;
			return false;
		}
		reset(static_cast<unsigned int>(neighborhoodSizes.size()));
		return true;
	}

	// Test the weighted k-NN classifier on all available parameters.
	bool DSDATest::test(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		string strMode = "";
		string errorTab = "";
		if (locallyRawFlag) 
			strMode = "locallyRaw";
		else if (locallySmoothFlag) 
			strMode = "locallySmooth";
		else if (PolyaFlag)
			strMode = "Polya";
		else if (PolyaSmoothedFlag)
			strMode = "PolyaSmoothed";
		else if (AllPairwiseFlag)
			strMode = "AllPairwise";
		else if (AllPairwiseLSDAFlag)
			strMode = "AllPairwiseLSDA";
		else
			cerr << "Regularization Scheme not defined" << endl;

		logFile << strMode << " Dirichlet Local SDA classifier" << endl
			<< "Neighborhood size\tError Rate" << endl;

		unsigned int i = 0; // parameter set count
		vector<int> results(data.getTestSetSize());

		for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
			itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
		{
			DSDA sda;
			sda.setNeighborhoodSize(*itrNeighborhoodSize);
			sda.setFlag(locallyRawFlag, locallySmoothFlag, PolyaFlag, PolyaSmoothedFlag, AllPairwiseFlag, AllPairwiseLSDAFlag);

			// Configure Dirichlet.
			if (!sda.configure(data)) return false;
			// Use Dirichlet to classify points.
			if (!sda.classify(data, results)) return false;

			// Calculate error rate.
			double errorRate = data.getClassificationError(results);
			// Log the error rate
			logFile << *itrNeighborhoodSize << "\t"
				<< (errorRate * 100.0) << "%" << endl;
			errorSum(i) += errorRate;
			errorSumOfSquares(i++) += (errorRate * errorRate);

		}

		logFile << endl;

		++testCount;

		return true;
	}

	bool DSDATest::setFlags(string mode, bool &locallyRawFlag, bool &locallySmoothFlag, bool &PolyaFlag, bool &PolyaSmoothedFlag, bool &AllPairwiseFlag, bool &AllPairwiseLSDAFlag)
{	
		// In DSDS, the mode should always be Dirichlet.  This code is leftover.
		locallyRawFlag = false;
		locallySmoothFlag = false;
		PolyaFlag = false;
		PolyaSmoothedFlag = false;
		AllPairwiseFlag = false;
		AllPairwiseLSDAFlag = false;

		if (mode == "locallyRaw")
			locallyRawFlag = true;
		else if (mode == "locallySmooth")
			locallySmoothFlag = true;
		else if (mode == "Polya")
			PolyaFlag = true;
		else if (mode == "PolyaSmoothed")
			PolyaSmoothedFlag = true;
		else if (mode == "AllPairwise")
			AllPairwiseFlag = true;
		else if (mode == "AllPairwiseLSDA")
			AllPairwiseLSDAFlag = true;
		else
			return false;
		return true;
	}

	// Check the legitimacy of the neighborhood sizes
	bool DSDATest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());
		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void DSDATest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void DSDATest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet
		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);
			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize = neighborhoodSizes[i];
		}
		resultsBestPara.push_back(latestResult);
	}

	// Test the weighted k-NN classifier with the best parameters
	bool DSDATest::testBestPara(const SimData &data, ofstream &logFile)
	{
		string mode = "";
		if (locallyRawFlag)
			mode = "locallyRaw";
		else if (locallySmoothFlag)
			mode = "locallySmooth";
		else if (PolyaFlag)
			mode = "Polya";
		else if (PolyaSmoothedFlag)
			mode = "PolyaSmoothed";
		else if (AllPairwiseFlag)
			mode = "AllPairwise";
		else if (AllPairwiseLSDAFlag)
			mode = "AllPairwiseLSDA";
		else
			return false;	
		logFile << mode << " Dirichlet Local SDA classifier" << endl;

		DSDA sda;
		unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		sda.setNeighborhoodSize(k);
		sda.setFlag(locallyRawFlag, locallySmoothFlag, PolyaFlag, PolyaSmoothedFlag, AllPairwiseFlag, AllPairwiseLSDAFlag);

		// Setup the Dirichlet.
		sda.configure(data);

		// Use Dirichlet to classify.
		vector<int> results(data.getTestSetSize());
		RealMatrix probMatrix;
		if (!sda.classify(data, results, probMatrix)) return false;

		// Record probabilities.
		logFile << "Classification probabilities:" << endl;
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				logFile << probMatrix(i,j) << " ";
			}
			logFile << endl;
		}
		logFile << "End of classification probabilities." << endl;
		// Calculate error rate.
		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;
		// Log the error rate
		logFile << "Neighborhood size\tLambda\tError rate" << endl
				<< k << "\t" << (errorRate * 100.0) << "%" << endl << endl;

		cout << "Finished another test." << endl; // PJS: probably remove later.
		return true;
	}

	// Report the test results of all available parameters
	void DSDATest::reportResults(ofstream &file, bool verbose)
	{
		string mode = "";
		if (locallyRawFlag)
			mode = "locallyRaw";
		else if (locallySmoothFlag)
			mode = "locallySmooth";
		else if (PolyaFlag)
			mode = "Polya";
		else if (PolyaSmoothedFlag)
			mode = "PolyaSmoothed";
		else if (AllPairwiseFlag)
			mode = "AllPairwise";
		else if (AllPairwiseLSDAFlag)
			mode = "AllPairwiseLSDAFlag";
		else
			assert(false);
			
		file << mode << " Dirichlet Local SDA classifier" << endl;
		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "Neighborhood size\tError rate\tStandard deviation" << endl;
			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				file << *itrNeighborhoodSize << "\t" 
					<< (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
				++i;

			}
		}
		else
		{
			// Report the best result only
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}
		file << endl;
	}

	// Report the test results with the best parameters
	void DSDATest::reportBestParaResults(ofstream &file, bool verbose)
	{
		string mode = "";
		if (locallyRawFlag)
			mode = "locallyRaw";
		else if (locallySmoothFlag)
			mode = "locallySmooth";
		else if (PolyaFlag)
			mode = "Polya";
		else if (PolyaSmoothedFlag)
			mode = "PolyaSmoothed";
		else if (AllPairwiseFlag)
			mode = "AllPairwise";
		else if (AllPairwiseLSDAFlag)
			mode = "AllPairwiseLSDA";
		else
			assert(false);	
			
		file << mode << " Dirichlet Local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void DSDATest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		string mode = "";
		if (locallyRawFlag)
			mode = "locallyRaw";
		else if (locallySmoothFlag)
			mode = "locallySmooth";
		else if (PolyaFlag)
			mode = "Polya";
		else if (PolyaSmoothedFlag)
			mode = "PolyaSmoothed";
		else if (AllPairwiseFlag)
			mode = "AllPairwise";
		else if (AllPairwiseLSDAFlag)
			mode = "AllPairwiseLSDA";
		else
			assert(false);	
			
		file << mode << " Dirichlet local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size\tError rate" << endl;
		}
		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->neighborhoodSize << "\t" 
					 << (itr->testError * 100.0) << "%" << endl;
			}
		}
		
		file << "Average neighborhood size: " << kVec.mean()
			 << " (" << kVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
