// File: LimvSolver.cpp
// Purpose: Implementation of the LimvSolver class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 1/31/2008 1:17:40 PM

#include <cstdio>
#include <iostream>
#include "mosek.h" // MOSEK header file
#include "LimvSolver.h"

using namespace std;

namespace idl
{
	// Initialize the number of equality constraints
	const int LimvSolver::NUM_EQ_CON = 1;

	MSKboundkeye LimvSolver::bkc[1] = {MSK_BK_FX}; // Type of constraint: fixed (equality)
	double LimvSolver::blc[1] = {1.0}; // Lower bound for the equality constraint
	double LimvSolver::buc[1] = {1.0}; // Upper bound for the equality constraint

	int *LimvSolver::ptrb = NULL;
	int *LimvSolver::ptre = NULL;
	int *LimvSolver::asub = NULL;
	double *LimvSolver::aval = NULL;

	int *LimvSolver::qsubi = NULL;
	int *LimvSolver::qsubj = NULL;
	double *LimvSolver::qval = NULL;

	MSKboundkeye *LimvSolver::bkx  = NULL;
	double *LimvSolver::blx = NULL;
	double *LimvSolver::bux = NULL;

	int LimvSolver::numVariables = 0;
	int LimvSolver::numQNz = 0;

	// Print the messages from MOSEK on the screen
	// Remarks: This is a callback function for MOSEK.
	void MSKAPI LimvSolver::mosekPrintLog(void *handle, char str[])
	{
#ifdef _DEBUG
		printf("%s", str);
#endif // _DEBUG
	}

	// Clear the information about the optimization problem
	void LimvSolver::clearProbInfo()
	{
		numVariables = 0;

		delete [] ptrb;
		delete [] ptre;
		delete [] asub;
		delete [] aval;
		delete [] bkx;
		delete [] blx;
		delete [] bux;
		delete [] qsubi;
		delete [] qsubj;
		delete [] qval;
	}

	// Set the information about the optimization problem
	bool LimvSolver::setProbInfo(int numNeighbors)
	{
		if (numNeighbors < 0)
		{
			cerr << "Error: Number of neighbors must be positive" << endl;
			return false;
		}

		if (numVariables != numNeighbors)
		{
			clearProbInfo();

			numVariables = numNeighbors;
			numQNz = numVariables * (numVariables + 1) / 2;

			ptrb = new int[numVariables];
			ptre = new int[numVariables];
			asub = new int[numVariables];
			aval = new double[numVariables];

			bkx = new MSKboundkeye[numVariables];
			blx = new double[numVariables];
			bux = new double[numVariables];
			
			qsubi = new int[numQNz];
			qsubj = new int[numQNz];
			qval  = new double[numQNz];

			int k = 0;
			for (int i = 0; i < numVariables; ++i)
			{
				ptrb[i] = i;			// Pointer to the first element in columns of A
				ptre[i] = i+1;			// Pointer to the last element + 1 in columns of A
				asub[i] = 0;			// Constraint subscripts
				aval[i] = 1.0;			// Constraint values (elements of A)
				bkx[i]  = MSK_BK_LO;	// Type of constraint: lower bound
				blx[i]  = 0.0;			// Lower bound of the inequality constraint
				bux[i]  = MSK_INFINITY;	// Upper bound of the inequality constraint

				for (int j = 0; j <= i; ++j)
				{
					qsubi[k] = i;		// i subscripts for the quadratic term
					qsubj[k] = j;		// j subscripts for the quadratic term
					++k;
				}
			}
		}

		return true;
	}

	// Generate matrix Q: Q = P + lambda * I
	void LimvSolver::generateQ(const double *P, double lambda)
	{
		int k = -1;
		for (int i = 0; i < numVariables; ++i)
		{
			for (int j = 0; j < i; ++j)
			{
				qval[++k] = P[i * numVariables + j];
			}
			qval[++k] = P[i * numVariables + i] + lambda;
		}
	}

	// Solve the quadratic program for LIMV
	bool LimvSolver::solve(int		numNeighbors,
						   double	*PMtx,
						   double	*qVec,
						   double	lambdaVal,
						   double	*weights)
	{
		if (NULL == weights)
		{
			cerr << "Error: Invalid pointer to the output buffer" << endl;
			return false;
		}

		if (!setProbInfo(numNeighbors)) return false;

		generateQ(PMtx, lambdaVal); // Generate matrix Q: Q = P + lambda * I

		///////////////////////////////////////////////////////////////////////
		// Solve the optimization problem by MOSEK
		///////////////////////////////////////////////////////////////////////

		MSKenv_t env; // MOSEK environment

		// Create a new MOSEK environment
		if (MSK_RES_OK != MSK_makeenv(&env, NULL, NULL, NULL, NULL)) 
		{
			cerr << "Error: Cannot create a new MOSEK environment" << endl;
			return false;
		}
		else
		{
			// Link MOSEK environment log stream to the standard output
			MSK_linkfunctoenvstream(env, MSK_STREAM_LOG, NULL, mosekPrintLog);
		}

		// Initialize the MOSEK environment
		if (MSK_RES_OK != MSK_initenv(env))
		{
			cerr << "Error: Cannot initialize the MOSEK environment" << endl;
			return false;
		}

		MSKtask_t task; // MOSEK task

		// Create an optimization task
		if (MSK_RES_OK != MSK_maketask(env, NUM_EQ_CON, numVariables, &task))
		{
			cerr << "Error: Cannot create an optimization task" << endl;
			return false;
		}
		else
		{
			// Link MOSEK task log stream to the standard output
			MSK_linkfunctotaskstream(task, MSK_STREAM_LOG, NULL, mosekPrintLog);
		}

		// Set the linear part of the optimization problem
		if (MSK_RES_OK != MSK_inputdata(
			task,			// A MOSEK optimization task
			NUM_EQ_CON,		// Maximum number of constraints in the optimization task
			numVariables,	// Maximum number of variables in the optimization task
			NUM_EQ_CON,		// Number of constraints
			numVariables,	// Number of variables
			qVec,			// Linear term in the objective function
			0.0,			// Fixed term in the objective function
			ptrb,			// Pointer to the first element in columns of A
			ptre,			// Pointer to the last element + 1 in columns of A
			asub,			// Constraint subscripts
			aval,			// Constraint values (elements of A)
			bkc,			// Bound keys for the constraints
			blc,			// Lower bounds for the constraints
			buc,			// Upper bounds for the constraints
			bkx,			// Bound keys for the variables
			blx,			// Lower bounds for the variables
			bux				// Upper bounds for the variables
			))
		{
			cerr << "Error: Cannot set the linear part of the optimization problem" << endl;
			return false;
		}

		// Set the quadratic part of the objective function
		if (MSK_RES_OK != MSK_putqobj(task, numQNz, qsubi, qsubj, qval))
		{
			cerr << "Error: Cannot set the quadratic part of the objective function" << endl;
			return false;
		}

		// Solve the optimization problem
		if (MSK_RES_OK != MSK_optimize(task))
		{
			cerr << "Error: Cannot solve the optimization problem" << endl;
			return false;
		}

		// Get the solution
		MSK_getsolutionslice(task, MSK_SOL_ITR, MSK_SOL_ITEM_XX, 0, numVariables, weights);

#ifdef _DEBUG
		// Print a short summary of the solution
		MSK_solutionsummary(task, MSK_STREAM_LOG);

		// Print the solution
		cout << "sol =" << endl;
		for (int i = 0; i < numVariables; ++i)
		{
			cout << weights[i] << endl;
		}
#endif // _DEBUG

		///////////////////////////////////////////////////////////////////////////
		// Clean up the environment
		///////////////////////////////////////////////////////////////////////////

		MSK_deletetask(&task);
		MSK_deleteenv(&env);

		return true;
	}
}
