// File: LimvSolver.h
// Purpose: Definition of the LimvSolver class
// Remarks: LimeSolver will call MOSEK (http://www.mosek.com/) to solve
//			the following quadratic program (QP):
//
//				min  1/2 * w' * (P + lambda * I) * w + q' * w
//				s.t. w >= 0, 1' * w = 1
//
//			The standard form of a quadratic program specified by MOSEK is:
//
//				min  1/2 * x' * Q * x + c' * x
//				s.t. l_c <= A * x <= u_c
//					 l_x <= x <= u_x
//
//			Here we have:
//
//				Q = P + lambda * I, c = q, A = 1'
//				l_c = u_c = 1, l_x = 0, u_x = inf
//
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 1/31/2008 1:17:40 PM

#ifndef IDL_LIMVSOLVER_H
#define IDL_LIMVSOLVER_H

#include "mosek.h" // MOSEK header file

namespace idl
{
	/**
	 * Class for computing the LIMV weights.
	 * @remarks		LimvSolver will call MOSEK to solve the LIMV problem (a QP).
	 * @attention	LimvSolver is a singleton implemented by static members.
	 * @warning		No instance of the LimvSolver class is allowed to be created. 
	 * @see			<a href="http://www.mosek.com/">MOSEK</a>
	 */
	class LimvSolver
	{
	public:
		/**
		 * Solve the quadratic program for LIMV.
		 * @param[in]	numNeighbors	number of neighbors
		 * @param[in]	PMtx			pointer to matrix P
		 * @param[in]	qVec			pointer to vector q
		 * @param[in]	lambdaVal		lambda value
		 * @param[out]	weights			pointer to the buffer to store the results
		 * @return		true if success; false if failure
		 */
		static bool solve(int		numNeighbors,
						  double	*PMtx,
						  double	*qVec,
						  double	lambdaVal,
						  double	*weights);

		/**
		 * Print the messages from MOSEK on the screen.
		 * @warning		DO NOT call this function. As a callback function
		 *				for MOSEK, this function is supposed to be called
		 *				by MOSEK ONLY.
		 */
		static void MSKAPI mosekPrintLog(void *handle, char str[]);

	private:
		/**
		 * Constructor.
		 * @remarks		To make the constructor private is to prevent
		 *				the creation of any LimvSolver object.
		 */
		LimvSolver() {};

		/**
		 * Set the information about the optimization problem.
		 * @param[in]	numNeighbors	number of neighbors
		 * @return		true if success; false if failure
		 */
		static bool setProbInfo(int numNeighbors);

		/** Clear the information about the optimization problem. */
		static void clearProbInfo();

		/**
		 * Generate matrix Q: Q = P + lambda * I.
		 * @param[in]	P		pointer to the buffer storing matrix P
		 * @param[in]	lambda	value of the regularization parameter
		 */
		static void generateQ(const double *P, double lambda);

	private:
		
		static const int NUM_EQ_CON; /**< Number of equality constraints. */

		static MSKboundkeye bkc[1]; /**< Bound keys for the constraints. */
		static double blc[1];		/**< Lower bounds for the constraints. */
		static double buc[1];		/**< Upper bounds for the constraints. */

		static int *ptrb;		/**< Start positions of the constraint values. */
		static int *ptre;		/**< End positions of the constraint values. */
		static int *asub;		/**< Constraint subscripts. */
		static double *aval;	/**< Constraint values. */

		static int *qsubi;		/**< i subscripts for the quadratic term. */
		static int *qsubj;		/**< j subscripts for the quadratic term. */
		static double *qval;	/**< Values of the quadratic term. */

		static MSKboundkeye *bkx;	/**< Bound keys for the variables. */
		static double *blx;			/**< Lower bounds for the variables. */
		static double *bux;			/**< Upper bounds for the variables. */

		static int numVariables;	/**< Number of varialbes. */
		static int numQNz;			/**< Number of non-zero elements in Q. */
	};
}

#endif // IDL_LIMVSOLVER_H
