// File: LocalSDA.cpp
// Purpose: Implementation of the LocalSDA class
// Remarks: This class is designed for LocalSDA classification.
// Author: Ingrid Liu (ingrid@hkn.eecs.berkeley.edu)
// Modified by Peter Sadowski peterjsadowski@gmail.com

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "LocalSDA.h"
#include "RealMatrix.h"
#include "LocalSDAModel.h"

using namespace std;

namespace idl
{
	// Classify the test samples of the similarity data set
	bool LocalSDA::classify(const SimData &data, vector<int> &predictedClassLabels, RealMatrix &classProbMatrix)
	{
		vector<int> uniqueClassLabels;
		data.getUniqueLabels(uniqueClassLabels);

		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		if (data.getTrainingSetSize() < parameters.neighborhoodSize)
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();
		predictedClassLabels.resize(numTestSamples);
		classProbMatrix.zeros(numTestSamples,(unsigned int)uniqueClassLabels.size());
		if (1 == parameters.neighborhoodSize)
		{
			// 1-NN
			for (unsigned int testIdx = 0; testIdx < numTestSamples; ++testIdx)
			{
				int label = data.get1NNLabel(testIdx);	
				for (unsigned int labelIdx = 0; labelIdx < (unsigned int)uniqueClassLabels.size(); ++labelIdx)
					if (label == uniqueClassLabels[labelIdx])
						classProbMatrix(testIdx,labelIdx) = 1;
				predictedClassLabels[testIdx] = label;
			}
			
		}
		else
		{
			
			RealMatrix simMat(parameters.neighborhoodSize,
							  parameters.neighborhoodSize);
			vector<int> sampleLabels(parameters.neighborhoodSize);
			RealMatrix simVec(1, parameters.neighborhoodSize);
			RealMatrix simDomain;
			data.getSimDomain(simDomain);
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				// Compute a local SDA model for this test point.
				LocalSDAModel lsdaModel;
				data.getNNLabels(i, parameters.neighborhoodSize, sampleLabels);
				data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
				data.getNNSimMat(i, parameters.neighborhoodSize, simMat);
				lsdaModel.initialize(simMat, sampleLabels, simVec, simDomain);
				// Classify and get class probabilities.
				map<int, double> tempProbs;
				predictedClassLabels[i] = lsdaModel.classify(simMat, sampleLabels, simVec, simDomain, tempProbs); 
				for (unsigned int labelIdx = 0; labelIdx < uniqueClassLabels.size(); ++labelIdx)
				{
					if (tempProbs.count(uniqueClassLabels[labelIdx]))
					{
						classProbMatrix(i,labelIdx) = tempProbs[uniqueClassLabels[labelIdx]];
					}
				}
				if (classProbMatrix.sumRow(i) < .99999 || classProbMatrix.sumRow(i) > 1.0001) 
				{
					std::cout << "Error: Probabilities summing to " << classProbMatrix.sumRow(i) << endl;
					return false;
				}
			}			
		}
		return true;
	}

	// Classify the test samples of the similarity data set and throw away the probability info.
	bool LocalSDA::classify(const SimData &data, vector<int> &predictedClassLabels)
	{
		RealMatrix tempClassProbMatrix;
		if (!classify(data, predictedClassLabels, tempClassProbMatrix)) {
			return false;
		}
		return true;
	}

}
