// File: LocalSDAModel.h
// Purpose: Definition of the LocalSDAModel class
// Author: Ingrid(ingrid@hkn.eecs.berkeley.edu), Peter(peterjsadowski@gmail.com)
// Last modified: 

// All indexed labels are "unsigned ints" - indexed from 0 .. |labels|
// All original labels are "ints"

#ifndef IDL_LOCALSDAMODEL_H
#define IDL_LOCALSDAMODEL_H

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include <vector>
#include <map>

namespace idl {

	double meanConstraintFxn(double lambda, const RealMatrix &domainVector, double classMean);
	// Calculate lambda and gamma for discrete exponential estimate.
	bool getLambdaGamma(const RealMatrix &domainVector, double classMean, double &lambda, double &gamma);
	// Calculate discrete exponential estimate pmf from a "raw" pmf using the mean sim.
	// Handles degenerate cases nicely.
	RealMatrix getDiscreteExponential(const RealMatrix &rawPMF, const RealMatrix &domainVector);

	class LocalSDAModel
	{
	
	public:

		LocalSDAModel();

		~LocalSDAModel();

		// Initialization of some basic and useful data structures.
		void initialize(const RealMatrix &simMat, const std::vector<int> &labels, const RealMatrix &simTest, const RealMatrix &domainVector);

		/* Compute the model parameter matrix.
		 * Since this takes time, only compute if we don't have a simple degenerate case.
		 */
		void computeModelParams(const RealMatrix &simMat, const std::vector<int> &labels,
			const RealMatrix &simTest, const RealMatrix &domainVector);

		double getSimTestToH(const RealMatrix &testSimVec,  int h);

		double getPrior(int g);

		// Returns likelihood P(s to each centroid | Y=G).
		double getNBLikelihoodGivenClass(const RealMatrix &testSimVec, const RealMatrix &domainVector, int g);

		/* Return the probability P(s(X, centroidH) | Y = G) 
		 * Used in RegLocalSDA. */
		double getSimProbGivenClass(RealMatrix &testSimVec, int g, int h, const RealMatrix &domainVector);

		/**
		 * Get row matrix of similarities from centroid of class h to samples in class g.
		 * @param[in]	simMat	
		 * @param[in]	g
		 * @param[in]	h
		 * @param[out]	HCenterToGSims	row matrix of similarities
		 * @return		true if success; false if failure
		 */		
		void getHCenterToGSims(const RealMatrix &simMat, RealMatrix &HCenterToGSims, int g, int h);

		/**
		 * Vector of sample indices from the specified class g. 
		 * @param[out]	sampleIndices	Vector of sample indices (w.r.t. simMat) of class g.  
		 * @param[in]	g				Class label.
		 */
		void getSamplesByLabel(std::vector<unsigned int> &sampleIndices, int g);


		// Copy of the vector containing the unique class labels that appear locally.
		void getUniqueLabels(std::vector<int> &uniqueLabels);

		/* Return the log posterior (Prior*Likelihood).
		 * P(Y=G | sim data for test point) = P(Y=G) * P(sim data | Y=G) */
		double posteriorFxn(const RealMatrix &simTest, const RealMatrix &simDomain, int predictedClass);

		int classify(RealMatrix &simMat, const std::vector<int> &orig_labels,
					 const RealMatrix &simTest, const RealMatrix &domainVector);

		int classify(RealMatrix &simMat, const std::vector<int> &orig_labels,
					 const RealMatrix &simTest, const RealMatrix &domainVector,
					 std::map<int, double> &probabilities);

		// Returns true if there is reason to use the backup classifier.
		bool haveReasonToUseBackupClassifier() {
			return  (anyOneClassFlag || onlyOneClassFlag || oneNeighbor || allSimsZeroFlag);}

		// Backup classifier for when local SDA fails.
		bool backupClassify(RealMatrix &simMat, const std::vector<int> &orig_labels,
						   const RealMatrix &simTest, const RealMatrix &domainVector,
						   int &predictedClass, std::map<int,double> &classProbs);

	private:
		
		//Initialize data structures

		/**
		 * Initialize samplesByLabel
		 * @param[in]	samplesByLabel	map from indexed label to indexed sample indices
		 * @param[in]	nSamples		# samples
		 * @return		true if success; false if failure
		 */		
		bool setSamplesByLabel(std::map<int,std::vector<unsigned int> > &samplesByLabel, const std::vector<int> orig_labels);
		     
		/**
		 * Initialize centroidIdx
		 * @param[in]	simMat			similarity matrix of samples
		 * @param[in]	centroidIdx		map from indexed label to indexed idx of centroid
		 * @return		true if success; false if failure
		 */		
		bool setClassCentroids(const RealMatrix &simMat, std::map<int,unsigned int> &centroidIdx);

		/**
		 * Initialize uniqueLabels, actualToIndexedLabels, indexedToActualLabels, indexedLabelsIdx
		 * @param[in]	labels					vector of original labels of samples
		 * @param[in]	uniqueClassLabelsLocal	map from indexed labels to original labels 
		 * @param[in]	actualToIndexedLabels	map from original labels to indexed labels
		 * @return		true if success; false if failure
		 */		
		bool setUniqueLabels(const std::vector<int> &labels, 
							 std::vector<int> &uniqueClassLabelsLocal, 
							 std::map<int,unsigned int> &actualToIndexedLabels);

		/**
		 * Initialize centroidsToTestSim
		 * @param[in]	centroidsToTestSim		1xC matrix similarity of centroids to test sample
		 * @param[in]	testSim					similarities of samples to test sample
		 * @return		true if success; false if failure
		 */	
		bool setCentroidsToTestSim(RealMatrix &centroidsToTestSim, const RealMatrix &testSim);
		
		/**
		 * Initialize classPriors
		 * @param[in]	classPriors				map from indexed labels to their prior probability
		 * @param[in]	nSamples				# samples
		 * @return		true if success; false if failure
		 */
		bool setPriors(std::map<int,double> &classPriors,int nSamples);

		/**
		 * Initialize anyOneClassFlag, onlyOneClassFlag, oneNeighbor, allSimsZeroFlag
		 * @param[in]	simTest					1*N vector of local sim values. 
		 * @param[in]	anyOneClassFlag			indicates whether there are any labels with only one sample
		 * @param[in]	onlyOneClassFlag		indicates whether all samples are from one class
		 * @param[in]	oneNeighbor				indicates whether the neighborhood size is 1
		 * @param[in]	allSimsZeroFlag				indicates whether all the local sims are zero, which indicates 
		 *										that we have no info on this sample (because all sims >= 0).
		 * @return		true if success; false if failure
		 */
		bool setDegenerateFlags(const RealMatrix &simTest, 
								bool &anyOneClassFlag, bool &onlyOneClassFlag,
								bool &oneNeighbor, bool &allSimsZeroFlag);

		bool reset();

		// (Used to be Protected)
		double sumSims(const RealMatrix &simMat, std::vector<unsigned int> &vect, unsigned int idx);

		// Convert class-label-to-weight map to class-label-to-prob map.
		bool computeProbsFromClassWeights(std::map<int, double> &classWeights,
										  std::map<int, double> &probabilities);

		// Data Structures for the Local SDA Model

		/**
		 * samplesByLabel (map)
		 * Key: int label
		 * Value: unsigned int vector, indexed indices of samples of indexed class Key
		 */
		std::map<int,std::vector<unsigned int> > samplesByLabel;

		/**
		 * uniqueClassLabelsLocal (vector<int>)
		 * Values: unique labels
		 */
		std::vector<int> uniqueClassLabelsLocal;

		/**
		 * actualToIndexedLabels (std::map<int,unsigned int>)
		 * Key: original label
		 * Value: indexed label
		 */
		std::map<int,unsigned int> actualToIndexedLabels;
		
		/**
		 * centroidIdx (map<int,unsigned int>)
		 * Key: label of the centroid
		 * Value: index of the centroid of label Key
		 */
		std::map<int,unsigned int> centroidIdx;

		/**
		 * centroidsToTestSim (RealMatrix)
		 * 1xC matrix of centroid similarity to test sample
		 * C = # centroids
		 */
		RealMatrix centroidsToTestSim;

		/**
		 * lambdas (RealMatrix)
		 * LxL matrix of lambda values
		 * lambdas(g,h) = mean of exponential model:
		 *   P(s(q,u_h(z)|Y=g)
		 */
		RealMatrix lambdas;

		/**
		 * gammas (RealMatrix)
		 * LxL matrix of gamma values
		 * gammas(g,h) = normalization constant of exponential model:
		 *   P(s(q,u_h(z)|Y=g)
		 */
		RealMatrix gammas;

		/**
		 * classPriors (std::map<int, double>)
		 * Key: indexed label
		 * Value: Laplace smoothed prior for label Key
		 */
		std::map<int, double> classPriors;

		/**
		 * anyOneClassFlag (bool)
		 * indicator for whether any sample is the sole sample of its label
		 */
		bool anyOneClassFlag;

		/**
		 * onlyOneClassFlag (bool)
		 * indicator for whether all samples are of the same label 
		 */
		bool onlyOneClassFlag;

		/**
		 * oneNeighbor (bool)
		 * indicator for whether the neighborhoodSize = 1
		 */
		bool oneNeighbor;

		/**
		 * allSimsZeroFlag (bool)
		 * Indicator for when all neighbors have similarity 0.
		 * We assume that we have no information about the sample and classify it
		 * using the global prior.
		 */
		bool allSimsZeroFlag;

	};
}	
#endif

