// File: NearestMean.h
// Purpose: Definition of the NearestMean class
// Remarks: This class is designed for nearest-mean classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 12/25/2007 11:59:22 PM

#ifndef IDL_NEARESTMEAN_H
#define IDL_NEARESTMEAN_H

#include <vector>
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"

namespace idl
{
	/**
	 * Class for the nearest-mean classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class NearestMean : public Classifier
	{
	public:
		/** Structure type for the nearest-mean classifier parameters. */
		struct ParaType
		{
			bool globalFlag;				/**< Flag for global/local mode. */
			unsigned int neighborhoodSize;	/**< Neighborhood size for each class. */
		};

		/** Default constructor. */
		NearestMean()
		{
			parameters.globalFlag = true;
			parameters.neighborhoodSize = 0;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the nearest-mean classifier
		 */
		NearestMean(ParaType para)
		{
			parameters = para;
		}

		~NearestMean() {}	/**< Destructor. */

		/**
		 * Set the flag for global/local mode.
		 * @param[in]	flag	flag for global/local mode
		 */
		void setGlobalFlag(bool flag)
		{
			parameters.globalFlag = flag;
		}

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSize(unsigned int k)
		{
			if (k > 0)
			{
				parameters.neighborhoodSize = k;
			}
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);

	private:
		/**
		 * Vote according to the mean similarity of each class.
		 * @param[in]	meanSim			mean similarity of each class
		 * @param[in]	classLabels		label of each class
		 * @return		estimated class label of the test sample
		 */
		static int vote(const std::vector<double> &meanSim, const std::vector<int> &classLabels);

	private:
		ParaType parameters;	/**< Parameters for the nearest-mean classifier. */
	};
}

#endif // IDL_NEARESTMEAN_H
