// File: PSVMTest.cpp
// Purpose: Implementation of the PSVMTest class
// Remarks: This class is designed as an adapter for the PSVM class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/18/2008 10:23:00 PM

#include <iostream>
#include <algorithm>
#include "utilities.h"
#include "PSVMTest.h"
#ifdef WINDOWS
#include <io.h>
#include <windows.h>
#else // LINUX
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#endif // WINDOWS

using namespace std;

namespace idl
{
	const string PSVMTest::_ID = "PSVM"; // Define the class ID

	// Register the class instantiator with the factory
	const bool PSVMTest::_REGISTERED =
		ClassifierTestFactory::registerCreator(PSVMTest::_ID, &PSVMTest::create);

	// Destructor
	PSVMTest::~PSVMTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Sort the elements of a real vector and check their positivity
	bool PSVMTest::checkPositive(vector<double> &vec)
	{
		sort(vec.begin(), vec.end());

		return vec.front() > 0.0;
	}

	// Configure the test for a PSVM classifier
	bool PSVMTest::configure(ConfigParser &config, int secIdx)
	{
		// Get the path of the PSVM binary
		string strPSVMBinary;
		if (!config.getString(secIdx, "Binary", strPSVMBinary))
		{
			cerr << "Error: Missing PSVM binary path for PSVM classifier " << secIdx << endl;
			return false;
		}
		// Check whether the executable exists or not
#ifdef WINDOWS
		if (0 != _access(strPSVMBinary.c_str(), 0))
#else // LINUX
		if (0 != access(strPSVMBinary.c_str(), X_OK))
#endif // WINDOWS
		{
			cerr << "Error: Bad PSVM binary path for PSVM classifier " << secIdx << endl;
			return false;
		}
		psvm.setBinaryPath(strPSVMBinary);

		// Get the working directory
		string strWorkingDirectory;
		if (!config.getString(secIdx, "WorkingDirectory", strWorkingDirectory))
		{
#ifdef ENABLE_WARNING
			cerr << "Warning: Missing working directory for PSVM classifier " << secIdx << endl;
#endif // ENABLE_WARNING
			// Use the default working directory
#ifdef WINDOWS
			const unsigned int bufferSize = 1024;
			char tmpFilePath[bufferSize];
			GetTempPath(bufferSize, tmpFilePath);
			strWorkingDirectory = tmpFilePath;
			strWorkingDirectory += "psvm\\";
#else // LINUX
			strWorkingDirectory = "/tmp/psvm/";
#endif // WINDOWS
		}
#ifdef WINDOWS
		if (0 != _access(strWorkingDirectory.c_str(), 0))
		{
			CreateDirectory(strWorkingDirectory.c_str(), NULL);
		}
		DWORD PID = GetCurrentProcessId();
		strWorkingDirectory += to_str(PID);
		strWorkingDirectory += '\\';
		CreateDirectory(strWorkingDirectory.c_str(), NULL);
#else // LINUX
		if (0 != access(strWorkingDirectory.c_str(), F_OK))
		{
			mkdir(strWorkingDirectory.c_str(), 0777);
		}
		pid_t PID = getpid();
		strWorkingDirectory += to_str(PID);
		strWorkingDirectory += '/';
		mkdir(strWorkingDirectory.c_str(), 0777);
#endif // WINDOWS
		psvm.setWorkingDirectory(strWorkingDirectory);

		// Get the hyperparameter C to cross-validate
		if (!config.getRealVec(secIdx, "C", Cs))
		{
			cerr << "Error: Missing or incorrect C for PSVM classifier " << secIdx << endl;
			return false;
		}
		if (!checkPositive(Cs))
		{
			cerr << "Error: Incorrect C for PSVM classifier " << secIdx << endl;
			return false;
		}

		// Get the hyperparameter epsilon to cross-validate
		if (!config.getRealVec(secIdx, "epsilon", epsilons))
		{
			cerr << "Error: Missing or incorrect epsilon for PSVM classifier " << secIdx << endl;
			return false;
		}
		if (!checkPositive(epsilons))
		{
			cerr << "Error: Incorrect epsilon for PSVM classifier " << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(Cs.size() * epsilons.size()));

		return true;
	}

	// Test the PSVM classifier on all available parameters
	bool PSVMTest::test(const SimData &data, ofstream &logFile)
	{
		logFile << "PSVM Classifier" << endl << "C\tepsilon\tError rate" << endl;

		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());
		
		for (vector<double>::const_iterator itr_C = Cs.begin(); itr_C != Cs.end(); ++itr_C)
		{
			psvm.setPara("C", *itr_C);

			for (vector<double>::const_iterator itr_epsilon = epsilons.begin();
				 itr_epsilon != epsilons.end(); ++itr_epsilon)
			{
				psvm.setPara("epsilon", *itr_epsilon);
				if (!psvm.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);

				// Log the error rate
				logFile << *itr_C << '\t' << *itr_epsilon << '\t'
						<< (errorRate * 100.0) << '%' << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			}
		}

		logFile << endl;

		++testCount;

		return true;
	}

	// Clear the results with the best parameters
	void PSVMTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void PSVMTest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.C = Cs.front();
			latestResult.epsilon = epsilons.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the C corresponding to the minimum cross-validation error
			latestResult.C = Cs[i / epsilons.size()];

			// Get the epsilon corresponding to the minimum cross-validation error
			latestResult.epsilon = epsilons[i % epsilons.size()];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the PSVM classifier with the best parameters
	bool PSVMTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		logFile << "PSVM Classifier" << endl;

		double C = (resultsBestPara.back()).C;
		psvm.setPara("C", C);

		double epsilon = (resultsBestPara.back()).epsilon;
		psvm.setPara("epsilon", epsilon);

		vector<int> results(data.getTestSetSize());
		if (!psvm.classify(data, results)) return false;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "C\tepsilon\tError rate" << endl
				<< C << '\t' << epsilon << '\t' << (errorRate * 100) << '%' << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void PSVMTest::reportResults(ofstream &file, bool verbose)
	{
		file << "PSVM Classifier" << endl;

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "C\tepsilon\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<double>::const_iterator itr_C = Cs.begin(); itr_C != Cs.end(); ++itr_C)
			{
				for (vector<double>::const_iterator itr_epsilon = epsilons.begin();
					 itr_epsilon != epsilons.end(); ++itr_epsilon)
				{
					file << *itr_C << '\t' << *itr_epsilon << '\t'
						 << (getError(i) * 100.0) << "%\t"
						 << (getErrorStdDev(i) * 100.0) << '%' << endl;
					++i;
				}
			}
		}
		else
		{
			// Report the best result only
			file << "C: " << (resultsBestPara.back()).C << endl
				 << "epsilon: " << (resultsBestPara.back()).epsilon << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << '%' << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << '%' << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void PSVMTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "PSVM Classifier" << endl
			 << "C: " << (resultsBestPara.back()).C << endl
			 << "epsilon: " << (resultsBestPara.back()).epsilon << endl;

		if (verbose)
		{
			file << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << '%' << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << '%' << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << '%' << endl;
		}
		else
		{
			file << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << '%' << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void PSVMTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "PSVM Classifier" << endl;

		if (verbose)
		{
			file << "C\tepsilon\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix CVec(1, n);
		RealMatrix epsilonVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			CVec(i) = itr->C;
			epsilonVec(i) = itr->epsilon;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->C << '\t' << itr->epsilon << '\t'
					 << (itr->testError * 100.0) << '%' << endl;
			}
		}

		file << "Average C: " << CVec.mean() << " (" << CVec.std() << ")" << endl
			 << "Average epsilon: " << epsilonVec.mean() << " (" << epsilonVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
