/*
 * C++ version of Minka's polya_fit.m MATLAB code from the fastfit toolbox.
 * By Peter Sadowski
 *
 *
 */

#include <iostream>
#include <math.h>
#include <cstdlib>
#include <float.h>
#include <vector>
#include "RealMatrix.h"
#include "UIntMatrix.h"
#include "util.h"
#include "PolyaFit.h"
#include <limits>

using namespace std;

namespace idl
{
	void polyaFit(const UIntMatrix &localHistMatrix, RealMatrix &a)
	{
		assert(localHistMatrix.min() >= 0);
		unsigned int nSimVals = localHistMatrix.getColumnNum();

		// Find sum of each column.
		UIntMatrix oneVec;
		oneVec.ones(1, localHistMatrix.getRowNum());
		UIntMatrix combinedHist = oneVec * localHistMatrix;
		// Sum the number of draws in each experiment. Call this rowSums.
		UIntMatrix oneColVec;
		oneColVec.ones(nSimVals, 1);
		UIntMatrix rowSums = localHistMatrix * oneColVec;

		// If we have no occurances of a sim value, we set its parameter to 0.
		if (combinedHist.min() == 0)
		{
			// Try to construct matrix where each sim value is seen at least once.
			vector<unsigned int> nonZeroIndices;
			for (unsigned int i = 0; i < nSimVals; ++i) 
			{
				if (combinedHist(i) > 0) {
					nonZeroIndices.push_back(i);
				}
			}
			// Set other parameters to 0.
			a.zeros(1, nSimVals);
			if (nonZeroIndices.size() > 0)
			{
				// Calculate alpha's for the sim values that are seen at least once.
				UIntMatrix subMatrix = localHistMatrix.columns(nonZeroIndices);
				RealMatrix subAlphaVec;
				// Find parameters for the relevant sim values.
				polyaFit(subMatrix, subAlphaVec);
				for (unsigned int i = 0; i < nonZeroIndices.size(); ++i) 
				{
					a(nonZeroIndices[i]) = subAlphaVec(i);
				}
			}
			return;
		}
		

		// Newton-Raphson

		
		if (a.length() == 0) {
			// Make an initial guess for the Polya parameters.
			RealMatrix guess;
			guess.zeros(1, nSimVals);
			for (unsigned int j = 0; j < nSimVals; ++j) {
				guess(j) = combinedHist(j);
			}
			a = guess;
		}
		double old_e = polyaLogProb(a, localHistMatrix).sum();
		// Iterate to improve this guess.
		double lambda = .1;
		vector<double> e;
		for (unsigned int iter = 0; iter < 100; ++iter)
		{
			if (a.sum() == 0) {
				break; }
			RealMatrix g = gradient2(a, localHistMatrix, rowSums);
			bool abort = false;
			// Newton iteration: Loop until we get a nonsingular hessian matrix.
			while (true)
			{
				RealMatrix hg = hessian_times_gradient2(a, localHistMatrix, rowSums, g, lambda);
				bool valid_hg = true;
				for (unsigned int i = 0; i < nSimVals; ++i) {
					if (hg(i) >= a(i)) {
						valid_hg = false;
						break;
					}
				}
				if (valid_hg) 
				{
					// Calculate new point.
					RealMatrix test_a = (a - hg);
					double test_e = polyaLogProb(test_a, localHistMatrix).sum();
					if (test_e > old_e)
					{
						// We've found a better test point.
						old_e = test_e;
						a -= hg;
						lambda = lambda/10;  // should this go in the if?
						e.push_back(test_e);
						break;
					}
				}
				// Increase lambda and try again.
				lambda *= 10;
				if (lambda > 1000000) {
					abort = true;
					break;
				}
			}
			// If search was aborted, end iterations.
			if (abort) {
				e.push_back(old_e);
				break;
			}
			// Keep all values in a numerically relevant.
			for (unsigned int i = 0; i < nSimVals; ++i) {
				if (a(i) < numeric_limits<double>::epsilon()) {
					a(i) = numeric_limits<double>::epsilon();
				}
			}
			// If gradient is small, end iterations.
			if ((g.max() < 1e-16) && (g.min() > -1e-16)) {
				break;
			}

		}
		return;
	}

	// Return a vector containing the log-probabilit of each histogram in data.
	// Follows Minka's polya_logProb.m.
	RealMatrix polyaLogProb(const RealMatrix &a, const UIntMatrix &localHistMatrix)
	{
		assert(a.getRowNum() == 1);  // If this assertion fails, see polya_logProb.m for fix.
		assert(a.min() >= 0);
		UIntMatrix temp;
		temp.ones(localHistMatrix.getColumnNum(), 1);
		UIntMatrix rowSum = localHistMatrix * temp;

		RealMatrix logp;
		logp.zeros(localHistMatrix.getRowNum(), 1);
		for (unsigned int i = 0; i < localHistMatrix.getColumnNum(); ++i)
		{
			for (unsigned int j = 0; j < localHistMatrix.getRowNum(); ++j)
			{
				logp(j) = logp(j) + pochhammer(a(i), localHistMatrix(j,i));
			}
		}
		for (unsigned int j = 0; j < localHistMatrix.getRowNum(); ++j)
		{
			logp(j) = logp(j) - pochhammer( a.sum(), rowSum(j));
		}
		return logp;
	}

	// Gradient.
	RealMatrix gradient2(const RealMatrix &a, const UIntMatrix &localHistMatrix, const UIntMatrix &rowSums)
	{
		unsigned int nSimVals = localHistMatrix.getColumnNum();
		unsigned int N = localHistMatrix.getRowNum();
		assert(nSimVals = a.getColumnNum());
		assert(N == rowSums.getRowNum());
		RealMatrix g;
		g.zeros(1, nSimVals);
		double sa = a.sum();
		double sum_di_pochhammer = 0;
		for (unsigned int i = 0; i < N; ++i) {
			for (unsigned int j = 0; j < nSimVals; ++j) {
				if (localHistMatrix(i,j) > 0) {
					g(j) += di_pochhammer(a(j), localHistMatrix(i,j));
				}
			}
			sum_di_pochhammer += di_pochhammer(sa, rowSums(i));
		}
		for (unsigned int j = 0; j < nSimVals; ++j) {
			g(j) -= sum_di_pochhammer;
			g(j) /= N;		// "For numeric stability."
		}
		return g;
	}

	// Hessian times gradient.
	RealMatrix hessian_times_gradient2(const RealMatrix &a, const UIntMatrix &localHistMatrix,
									   const UIntMatrix &rowSums, const RealMatrix &g,
									   const double lambda)
	{
		unsigned int nSimVals = localHistMatrix.getColumnNum();
		unsigned int N = localHistMatrix.getRowNum();
		assert(nSimVals = a.getColumnNum());
		assert(nSimVals = g.getColumnNum());
		assert(N == rowSums.getRowNum());
		double sa = a.sum();
		RealMatrix q;
		q.zeros(1, nSimVals);
		double sum_tri_pochhammer = 0;
		for (unsigned int i = 0; i < N; ++i) {
			for (unsigned int j = 0; j < nSimVals; ++j) {
				if (localHistMatrix(i,j) > 0) {
					q(j) += tri_pochhammer(a(j), localHistMatrix(i,j));
				}
			}
			sum_tri_pochhammer += tri_pochhammer(sa, rowSums(i));
		}
		double z = -sum_tri_pochhammer / N;
		q /= N;
		q -= lambda;
		double b = 0;
		for (unsigned int j = 0; j < nSimVals; ++j) {
			assert(q(j) != 0);
			q(j) = 1/q(j);
			b += (g(j) * q(j));
		}
		//double b = g.dot(q);
		b /= (1.0/z + q.sum());
		RealMatrix hg;
		hg.zeros(1,nSimVals);
		for (unsigned int j = 0; j < nSimVals; ++j) {
			hg(j) = (g(j) - b) * q(j);		
		}
		return hg;
	}

}

