// File: RegLocalSDABN.h
// Purpose: Definition of the RegLocalSDABN class
// Remarks: This class is designed for Regularized Local SDA classification.
// Author: Ingrid(ingrid@hkn.eecs.berkeley.edu)
// Last modified: 7/11/2009 11:45:18 AM

#ifndef IDL_RegLocalSDABN_H
#define IDL_RegLocalSDABN_H

#include <vector>
#include <fstream>
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "LocalSDAModel.h"
#include "RegLocalSDA.h"

namespace idl
{
	//double meanConstraintFxn(double lambda, RealMatrix &domainVector, double classMean);

	/**
	 * Class for the Local SDA classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class RegLocalSDABN : public RegLocalSDA
	{
	public:
		/** Enumeration type for embedding method. */

		/** Structure type for the Local SDA parameters. */
		struct ParaType
		{
			std::vector<unsigned int> neighborhoodSizes;	/**< Neighborhood sizes. */
			double regParam;								/**< Regularization parameter  */
		};

		/** Default constructor. */
		RegLocalSDABN()
		{
			parameters.regParam = -1.0;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the Local SDA classifier
		 */
		RegLocalSDABN(const ParaType &para)
		{
			parameters = para;
		}

		~RegLocalSDABN() {}	/**< Destructor. */

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSizes(std::vector<unsigned int> nBSizes)
		{
				parameters.neighborhoodSizes = nBSizes;
		}

		/**
		 * Set the regularization parameter.
		 * @param[in]	regParam	regularization parameter regParam for l-SDA
		 */
		void setRegPara(double regParam)
		{
			parameters.regParam = regParam;
		}

		/**
		 * Set the parameters for the Local SDA classifier.
		 * @param[in]	para	parameters for the Local SDA classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);
		bool classify(const SimData &data, std::vector<int> &results, bool jointFlag);
		bool classify(const SimData &data, std::vector<int> &results, bool jointFlag,RealMatrix &probMatrix);

		//inherited
		//bool configure(const SimData &data);

	private:
		ParaType parameters;	/**< Parameters for the Local SDA classifier. */
		int maxLabel(const std::vector<int> &labels, const RealMatrix &probabilities);

	};
}

#endif // IDL_LocalSDA_H
