// File: SVM.cpp
// Purpose: Implementation of the SVM class
// Remarks: This class is designed for SVM classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/19/2008 11:34:33 AM

#include <iostream>
#include <algorithm>
#include "SVM.h"

using namespace std;

namespace idl
{
	// Initialize the string representations of the kernels
	const char *SVM::kernelNames[] = {
		"linear kernel",
		"polynomial kernel",
		"Gaussian radial basis function kernel",
		"sigmoid kernel",
		"similarity kernel with eigenvalue shift",
		"similarity kernel with eigenvalue clip",
		"similarity kernel with eigenvalue flip" };

	// Classify the test samples of the similarity data set
	bool SVM::classify(const SimData &data, vector<int> &results)
	{
		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		vector<int> trainingLabels;
		RealMatrix trainingSamples;
		RealMatrix testSample;

		if (KERNEL_SIM_SHIFT == parameters.kernel ||
			KERNEL_SIM_CLIP == parameters.kernel ||
			KERNEL_SIM_FLIP == parameters.kernel)
		{
			// Treat similarities as inner products
			RealMatrix simMat;
			RealMatrix simVec;
			RealMatrix embeddedSamples;

			if (data.isLOOCV())
			{
				for (unsigned int i = 0; i < numTestSamples; ++i)
				{
					data.getTrainingLabels(trainingLabels, i);

					if (parameters.isStrictlySupervised)
					{
						data.getTrainingSimMat(simMat, i);
						data.getTestSimVec(i, simVec);
						embedder.embed(simMat, trainingSamples);
						embedder.project(simVec, testSample);
					}
					else
					{
						data.getTrainingTestSimMat(simMat, i);
						embedder.embed(simMat, embeddedSamples);
						trainingSamples = embeddedSamples.rows(0, numTestSamples - 2);
						testSample = embeddedSamples.row(numTestSamples - 1);
					}

					if (!train(trainingSamples, trainingLabels)) return false;

					results[i] = predict(testSample);
				}
			}
			else
			{
				data.getTrainingLabels(trainingLabels);

				if (parameters.isStrictlySupervised)
				{
					data.getTrainingSimMat(simMat);
					embedder.embed(simMat, trainingSamples);

					if (!train(trainingSamples, trainingLabels)) return false;

					for (unsigned int i = 0; i < numTestSamples; ++i)
					{
						data.getTestSimVec(i, simVec);
						embedder.project(simVec, testSample);
						results[i] = predict(testSample);
					}
				}
				else
				{
					// Get the number of training samples
					unsigned int numTrainingSamples = data.getTrainingSetSize();

					data.getTrainingTestSimMat(simMat);
					embedder.embed(simMat, embeddedSamples);
					trainingSamples = embeddedSamples.rows(0, numTrainingSamples - 1);

					if (!train(trainingSamples, trainingLabels)) return false;

					for (unsigned int i = 0; i < numTestSamples; ++i)
					{
						testSample = embeddedSamples.row(numTrainingSamples + i);
						results[i] = predict(testSample);
					}
				}
			}
		}
		else // non-similarity kernels
		{
			// Treat similarities directly as features
			if (data.isLOOCV())
			{
				for (unsigned int i = 0; i < numTestSamples; ++i)
				{
					data.getTrainingSimMat(trainingSamples, i);
					data.getTrainingLabels(trainingLabels, i);

					if (!train(trainingSamples, trainingLabels)) return false;

					data.getTestSimVec(i, testSample);
					results[i] = predict(testSample);
				}
			}
			else
			{
				data.getTrainingSimMat(trainingSamples);
				data.getTrainingLabels(trainingLabels);

				if (!train(trainingSamples, trainingLabels)) return false;

				for (unsigned int i = 0; i < numTestSamples; ++i)
				{
					data.getTestSimVec(i, testSample);
					results[i] = predict(testSample);
				}
			}
		}

		return true;
	}

	// Train the SVM
	bool SVM::train(const RealMatrix &trainingSamples, const vector<int> &labels)
	{
		unsigned int n = trainingSamples.getRowNum();		// Number of the training samples
		unsigned int d = trainingSamples.getColumnNum();	// Dimension of the feature vectors

		CvMat *_trainingData = cvCreateMat(n, d, CV_32FC1);	// Training data
		for (unsigned int i = 0; i < n; ++i)
		{
			for (unsigned int j = 0; j < d; ++j)
			{
				cvmSet(_trainingData, i, j, static_cast<float>(trainingSamples(i, j)));
			}
		}

		CvMat *_labels = cvCreateMat(n, 1, CV_32SC1); // Class labels of the training data
		copy(labels.begin(), labels.end(), reinterpret_cast<int *>(cvPtr1D(_labels, 0)));

		// Set the SVM parameters
		CvSVMParams svmParams;
		svmParams.svm_type = CvSVM::C_SVC;	// C-SVM
		switch (parameters.kernel)
		{
		case KERNEL_LINEAR:		// Linear kernel: k(x, x') = <x, x'>
		case KERNEL_SIM_SHIFT:	// Similarity kernel with eigenvalue shift
		case KERNEL_SIM_CLIP:	// Similarity kernel with eigenvalue clip
		case KERNEL_SIM_FLIP:	// Similarity kernel with eigenvalue flip
			svmParams.kernel_type = CvSVM::LINEAR;
			break;
		case KERNEL_POLY: // Polynomial kernel: k(x, x') = (gamma * <x, x'> + r)^d
			svmParams.kernel_type = CvSVM::POLY;
			svmParams.gamma = parameters.gamma;
			svmParams.coef0 = parameters.r;
			svmParams.degree = parameters.d;
			break;
		case KERNEL_RBF: // Gaussian radial basis function kernel: k(x, x') = exp(-gamma * ||x - x'||^2)
			svmParams.kernel_type = CvSVM::RBF;
			svmParams.gamma = parameters.gamma;
			break;
		case KERNEL_SIGMOID: // Sigmoid kernel: k(x, x') = tanh(gamma * <x, x'> + r)
			svmParams.kernel_type = CvSVM::SIGMOID;
			svmParams.gamma = parameters.gamma;
			svmParams.coef0 = parameters.r;
			break;
		}
		svmParams.C = parameters.C; // Penalty parameter for C-SVM
		svmParams.class_weights = NULL;
		svmParams.term_crit = cvTermCriteria(CV_TERMCRIT_ITER + CV_TERMCRIT_EPS, 3000, 1e-8);

		// Train the C-SVM
		bool isOK = true;
		if (!svm.train(_trainingData, _labels, NULL, NULL, svmParams))
		{
			cerr << "Error: Cannot train an SVM" << endl;
			isOK = false;
		}

		cvReleaseMat(&_trainingData);
		cvReleaseMat(&_labels);

		return isOK;
	}

	// Predict the label of a test sample
	int SVM::predict(const RealMatrix &testSample)
	{
		unsigned int d = testSample.length();

		CvMat *_testSample = cvCreateMat(1, d, CV_32FC1); // Test sample
		for (unsigned int j = 0; j < d; ++j)
		{
			cvmSet(_testSample, 0, j, static_cast<float>(testSample(j)));
		}

		// Predict the class label of the test sample
		int result = static_cast<int>(svm.predict(_testSample));

		cvReleaseMat(&_testSample);

		return result;
	}
}
