// File: SVMKNN.h
// Purpose: Definition of the SVMKNN class
// Remarks: This class is designed for SVM-KNN classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 1/7/2008 4:51:37 PM

#ifndef IDL_SVMKNN_H
#define IDL_SVMKNN_H

#include <vector>
#include "cxcore.h"	// OpenCV CXCORE header file
#include "ml.h"		// OpenCV MLL header file
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "SimEmbed.h"

namespace idl
{
	/**
	 * Class for the SVM-KNN classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class SVMKNN : public Classifier
	{
	public:
		/** Enumeration type for embedding method. */
		enum EmbeddingType
		{
			EMBED_SHIFT,	/**< embedding with eigenvalue shift */
			EMBED_CLIP,		/**< embedding with eigenvalue clip */
			EMBED_FLIP,		/**< embedding with eigenvalue flip */
			EMBED_NONE		/**< no embedding */
		};

		/** Structure type for the SVM-KNN parameters. */
		struct ParaType
		{
			unsigned int neighborhoodSize;	/**< Neighborhood size. */
			EmbeddingType	embeddingMode;	/**< Embedding mode. */
			double C;						/**< Regularization parameter C for C-SVM. */
		};

		/** Default constructor. */
		SVMKNN()
		{
			parameters.neighborhoodSize = 1;
			parameters.embeddingMode = EMBED_SHIFT;
			parameters.C = 1.0;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the SVM-KNN classifier
		 */
		SVMKNN(const ParaType &para)
		{
			parameters = para;
			_setEmbedder();
		}

		~SVMKNN() {}	/**< Destructor. */

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSize(unsigned int k)
		{
			if (k > 0)
			{
				parameters.neighborhoodSize = k;
			}
		}

		/**
		 * Set the embedding mode.
		 * @param[in]	t	embedding mode
		 */
		void setEmbeddingMode(EmbeddingType t)
		{
			parameters.embeddingMode = t;
			_setEmbedder();
		}

		/**
		 * Set the regularization parameter C.
		 * @param[in]	C	regularization parameter C for C-SVM
		 */
		void setRegPara(double C)
		{
			parameters.C = C;
		}

		/**
		 * Set the parameters for the SVM-KNN classifier.
		 * @param[in]	para	parameters for the SVM-KNN classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
			_setEmbedder();
		}

		/**
		 * Get the embedding mode.
		 * @return	embedding mode
		 */
		EmbeddingType getEmbeddingMode() const
		{
			return parameters.embeddingMode;
		}

		/**
		 * Get the name of the embedding method.
		 * @return	string representation of the embedding method
		 */
		const char *getEmbeddingName() const
		{
			return embeddingNames[parameters.embeddingMode];
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);

	private:
		/**
		 * Check if the given labels belong to one class.
		 * @param[in]	labels		vector of the given class labels
		 * @return		true if the given labels belong to one class;
		 *				false if they belong to more than one class
		 */
		static bool isOneClass(const std::vector<int> &labels);

		/**
		 * Estimate the class label of a test sample.
		 * Estimate the class label of a test sample by building a local SVM
		 * from its neighbors.
		 * @param[in]	simMat		similarity matrix of the test sample and its neighbors
		 * @param[in]	labels		class labels of the neighbors
		 * @return		estimated class label of the test sample
		 * @remarks		Let k denote the number of the neighbors. The similarity matrix is
		 *				(k + 1)-by-(k + 1) with the last row and column for the test sample.
		 */
		int estimateClassLabel(const RealMatrix &simMat, const std::vector<int> &labels);

		/** Set the embedding method of the embedder. */
		void _setEmbedder()
		{
			if (EMBED_SHIFT == parameters.embeddingMode)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_SHIFT);
			}
			else if (EMBED_CLIP == parameters.embeddingMode)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_CLIP);
			}
			else if (EMBED_FLIP == parameters.embeddingMode)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_FLIP);
			}
		}

	private:
		/** String representations of the embedding methods. */
		static const char *embeddingNames[];

		ParaType parameters;	/**< Parameters for the SVM-KNN classifier. */
		SimEmbed embedder;		/**< Embedder for embedding similarity data. */
		CvSVM svm;				/**< Multi-class SVM from OpenCV MLL. */
	};
}

#endif // IDL_SVMKNN_H
