// File: SVMTest.h
// Purpose: Definition of the SVMTest class
// Remarks: This class is designed as an adapter for the SVM class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:23:31 PM

#ifndef IDL_SVMTEST_H
#define IDL_SVMTEST_H

#include <vector>
#include <string>
#include <fstream>
#include "ClassifierTest.h"
#include "SVM.h"

namespace idl
{
	/**
	 * Class for testing an SVM classifier on similarity data.
	 * An adapter class for the SVM class to enable testing on multiple
	 * sets of parameters.
	 * @remarks	This class is derived from the ClassifierTest class.
	 * @see		ClassifierTest
	 * @see		ClassifierTestFactory
	 * @see		SVM
	 */
	class SVMTest : public ClassifierTest
	{
	public:
		SVMTest() : ClassifierTest() {}	/**< Constructor. */
		~SVMTest();						/**< Destructor. */

		/** Clear the results with the best parameters. */
		void resetBestParaResults();

		/** Select the best parameters based on the previous tests. */
		void selectBestPara();

		/**
		 * Configure the test for an SVM classifier.
		 * @param[in]	config		ConfigParser object containing user configurations
		 * @param[in]	secIdx		section index
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config, int secIdx);

		/**
		 * Test the SVM classifier on all available parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool test(const SimData &data, std::ofstream &logFile);

		/**
		 * Test the SVM classifier with the best parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool testBestPara(const SimData &data, std::ofstream &logFile);

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 */
		void reportResults(std::ofstream &file, bool verbose = false);

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaResults(std::ofstream &file, bool verbose = false);

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaSummary(std::ofstream &file, bool verbose = false);

	private:
		SVM svm;					/**< SVM classifier. */
		std::vector<double> Cs;		/**< Vector of penalty parameters for C-SVM. */
		std::vector<double> gammas;	/**< Vector of "gamma" parameters for polynomial, RBF or sigmoid kernel. */
		std::vector<double> rs;		/**< Vector of "r" parameters for polynomial or sigmoid kernel. */
		std::vector<double> ds;		/**< Vector of "d" parameters for polynomial kernel. */

		/** Structure type for test results with the best parameters. */
		struct ResultType
		{
			double C;				/**< Penalty parameter for C-SVM. */
			double gamma;			/**< Kernel parameter for polynomial, RBF or sigmoid kernel. */
			double r;				/**< Kernel parameter for polynomial or sigmoid kernel. */
			double d;				/**< Kernel parameter for polynomial kernel. */
			double cvError;			/**< Cross-validation error. */
			double cvErrorStdDev;	/**< Standard deviation of the cross-validation error. */
			double testError;		/**< Test error. */
		};

		std::vector<ResultType> resultsBestPara; /**< Test results with the best parameters. */

	private:
		/**
		 * Print the header information.
		 * @param[in]	file	output file stream
		 */
		void _printHeader(std::ofstream &file);

		/**
		 * Sort the elements of a real vector and check their positivity.
		 * @param[in, out]	vec		real vector to sort and check
		 * @return			true if all the elements are positive;
		 *					false if at least one element is not positive
		 */
		static bool checkPositive(std::vector<double> &vec);

		/**
		 * Create an object of the SVMTest class.
		 * @return	pointer to a new object of the SVMTest class
		 * @remarks	This is the creator function to be registered with the factory.
		 */
		static ClassifierTest *create()
		{
			return new SVMTest;
		}
		static const std::string _ID;	/**< Class ID ("SVM") */
		static const bool _REGISTERED;	/**< Flag for creator registration */
	};
}

#endif // IDL_SVMTEST_H
