// File: SimData.h
// Purpose: Definition of the SimData class
// Remarks: This class is designed for classification data.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 7/6/2008 1:09:23 AM

#ifndef IDL_SIMDATA_H
#define IDL_SIMDATA_H

#include <climits>
#include <string>
#include <vector>
#include <map>
#include <fstream>
#include "Matrix.h"
#include "RealMatrix.h"
#include "UIntMatrix.h"

namespace idl
{
	/** Class for storing and managing similarity data. */
	class SimData
	{
	public:
		/** Constructor. */
		SimData() : numSamples(0), numClasses(0), normalizationFlag(false), symmetryFlag(false), loocvFlag(true) {}
		~SimData() {} /**< Destructor. */

		void clear(); /**< Clear the current similarity data. */
		void reset(); /**< Reset the training and test settings. */

		/**
		 * Log data information.
		 * @param[in]	file	output file stream
		 */
		void logDataInfo(std::ofstream &file) const;

		/**
		 * Check if the data is normalized.
		 * @return		true if normalized; false if not normalized
		 */
		bool isNormalized() const
		{
			return normalizationFlag;
		}

		/**
		 * Check if the data is symmetric.
		 * @return		true if symmetric; false if asymmetric
		 */
		bool isSymmetric() const
		{
			return symmetryFlag;
		}

		/**
		 * Check if the data is for LOOCV.
		 * @return		true if the data is for LOOCV; false if the data is not for LOOCV
		 */
		bool isLOOCV() const
		{
			return loocvFlag;
		}

		/**
		 * Load similarity data and class labels.
		 * @param[in]	dataFileName	name of the data file
		 * @param[in]	dataFormat		format of the data file
		 * @param[in]	labelFileName	name of the label file
		 * @param[in]	symmetrization	flag for symmetrization (true to force symmetrization)
		 * @return		true if success; false if failure
		 */
		bool loadData(const std::string &dataFileName, const std::string &dataFormat,
					  const std::string &labelFileName, bool symmetrization);

		/**
		 * Normalize similarity data to the range of [0, 1].
		 * @return		true if success; false if failure
		 */
		bool normalizeData();

		/**
		 * Set the training and test set.
		 * @param[in]	trainingIdx	indices of the training samples
		 * @param[in]	testIdx		indices of the test samples
		 * @return		true if success; false if failure
		 */
		bool setTrainingTest(const std::vector<unsigned int> &trainingIdx,
							 const std::vector<unsigned int> &testIdx);

		/**
		 * Set a subset of the samples for leave-one-out cross-validation.
		 * @param[in]	testIdx	indices of the samples in the subset
		 * @return		true if success; false if failure
		 */
		bool setLoocvTest(const std::vector<unsigned int> &testIdx);

		/**
		 * Get the number of classes.
		 * @return	number of classes
		 */
		unsigned int getNumClasses() const
		{
			return numClasses;
		}

		/**
		 * Get the data set size.
		 * @return	number of samples in the data set
		 */
		unsigned int getDataSetSize() const
		{
			return numSamples;
		}

		/**
		 * Get the test set size.
		 * @return	number of samples in the test set
		 */
		unsigned int getTestSetSize() const
		{
			return testSet.empty() ? numSamples : testSet.size();
		}

		/**
		 * Get the training set size.
		 * @return	number of samples in the training set
		 */
		unsigned int getTrainingSetSize() const
		{
			return testSet.empty() ? (numSamples - 1)
				: (loocvFlag ? (testSet.size() - 1) : trainingSet.size());
		}

		/**
		 * Get the similarity matrix of the nearest neighbors.
		 * Get the similarities between the nearest neighbors of a test sample.
		 * @param[in]	testSampleIdx		index of the test sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnSimMat			retrieved similarity matrix
		 */
		void getNNSimMat(unsigned int testSampleIdx,
						 unsigned int neighborhoodSize,
						 RealMatrix   &nnSimMat) const;

		/**
		 * Get the similarity matrix of the nearest neighbors.
		 * Get the similarities between the nearest neighbors of any sample.
		 * @param[in]	sampleIdx			index of the sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnSimMat			retrieved similarity matrix
		 */
		void getNNofTrainingSimMat(unsigned int sampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimMat) const;

		/**
		 * Get the similarity matrix of the test sample and its nearest neighbors.
		 * @param[in]	testSampleIdx		index of the test sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnTestSimMat		retrieved similarity matrix
		 * @remarks		Let k denote the neighborhood size. The retrieved similarity
		 *				matrix is (k + 1)-by-(k + 1) with the last row and column for
		 *				the test sample.
		 */
		void getNNTestSimMat(unsigned int testSampleIdx,
							 unsigned int neighborhoodSize,
							 RealMatrix   &nnTestSimMat) const;

		/**
		 * Get the similarity vector to the nearest neighbors.
		 * Get the similarities between any sample and its nearest neighbors.
		 * @param[in]	sampleIdx		index of the sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnSimVec			retrieved similarity vector
		 */
		void getNNofTrainingSimVec(unsigned int sampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimVec) const;

		/**
		 * Get the similarity vector to the nearest neighbors.
		 * Get the similarities between the test sample and its nearest neighbors.
		 * @param[in]	testSampleIdx		index of the test sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnSimVec			retrieved similarity vector
		 */
		void getNNSimVec(unsigned int testSampleIdx,
						 unsigned int neighborhoodSize,
						 RealMatrix   &nnSimVec) const;

		/**
		 * Get the labels of the nearest neighbors.
		 * Get the labels of the nearest neighbors of any sample.
		 * @param[in]	sampleIdx		index of the test sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnLabels			retrieved label vector
		 */
		void getNNofTrainingLabels(unsigned int sampleIdx,
						 unsigned int neighborhoodSize,
						 std::vector<int> &nnLabels) const;

		/**
		 * Get the labels of the nearest neighbors.
		 * Get the labels of the nearest neighbors of a test sample.
		 * @param[in]	testSampleIdx		index of the test sample
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @param[out]	nnLabels			retrieved label vector
		 */
		void getNNLabels(unsigned int testSampleIdx,
						 unsigned int neighborhoodSize,
						 std::vector<int> &nnLabels) const;

		int get1NNofTrainingLabel(unsigned int sampleIdx) const;

		/**
		 * Get the label of the 1-nearest-neighbor.
		 * Get the label of the 1-nearest-neighbor of a test sample.
		 * @param[in]	testSampleIdx	index of the test sample
		 * @return		label of the 1-nearest-neighbor
		 */
		int  get1NNLabel(unsigned int testSampleIdx) const;

		/**
		 * Get the similarity matrix of the training samples.
		 * @param[out]	trainingSimMat	retrieved similarity matrix
		 * @param[in]	testSampleIdx	index of the test sample (needed only when the data is
		 *								for leave-one-out cross-validation)
		 */
		void getTrainingSimMat(RealMatrix &trainingSimMat,
							   unsigned int testSampleIdx = 0) const;

		/**
		 * Get the labels of the training samples.
		 * @param[out]	trainingLabels	retrieved label vector
		 * @param[in]	testSampleIdx	index of the test sample (needed only when the data is
		 *								for leave-one-out cross-validation)
		 */
		void getTrainingLabels(std::vector<int> &trainingLabels,
							   unsigned int testSampleIdx = 0) const;

		/**
		 * Get the similarity matrix of the training and test samples.
		 * @param[out]	trainingTestSimMat	retrieved similarity matrix
		 * @param[in]	testSampleIdx		index of the test sample (needed only when the data is
		 *									for leave-one-out cross-validation)
		 */
		void getTrainingTestSimMat(RealMatrix &trainingTestSimMat,
								   unsigned int testSampleIdx = 0) const;

		/**
		 * Get the similarity vector of the test sample to the training samples.
		 * Get the similarities between the test sample and the training samples.
		 * @param[in]	testSampleIdx	index of the test sample
		 * @param[out]	testSimVec		retrieved similarity vector
		 */
		void getTestSimVec(unsigned int testSampleIdx, RealMatrix &testSimVec) const;

		/**
		 * Get the similarities of the test samples to the training samples.
		 * Get the similarities between the test samples and the training samples.
		 * @param[out]	testSims	retrieved similarity vectors (row vectors)
		 * @attention	This function should not be called when the data is for
		 *				leave-one-out cross-validation.
		 */
		void getTestSims(RealMatrix &testSims) const;

		/**
		 * Get all unique discrete similarities in the similarity matrix.
		 * @param[out]	uniqueVals retrieved Nx1 RealMatrix of similarities
		 */
		void getSimDomain(RealMatrix &uniqueVals) const;

		/**
		 * Get all unique labels in the similarity matrix.
		 * @param[out]	uniqueLabels retrieved vector of labels
		 */
		void getUniqueLabels(std::vector<int> &uniqueLabels) const;

		/**
		 * Get the mean similarity of the test sample and it nearest neighbors of each class.
		 * @param[in]	testSampleIdx		index of the test sample
		 * @param[out]	meanSim				mean similarity of each class
		 * @param[out]	classLabels			label of each class
		 * @param[in]	neighborhoodSize	neighborhood size
		 * @remarks		All the training samples will be considered when the default value of
		 *				the neighborhood size is used.
		 */
		void getMeanSim(unsigned int testSampleIdx,
						std::vector<double> &meanSim,
						std::vector<int> &classLabels,
						unsigned int neighborhoodSize = UINT_MAX) const;

		/**
		 * Get the classification error.
		 * @param[in]	estimatedLabels		estimated class labels
		 * @return		classification error (in the range of [0, 1])
		 */
		double getClassificationError(std::vector<int> &estimatedLabels) const;

		/**
		 * Get the classification error and log the classification results.
		 * @param[in]	estimatedLabels		estimated class labels
		 * @param[in]	logFile				log file stream
		 * @return		classification error (in the range of [0, 1])
		 */
		double getClassificationError(std::vector<int> &estimatedLabels, std::ofstream &logFile) const;

	private:
		/**
		 * Read similarity data from file.
		 * @param[in]	fileName	data file name
		 * @param[in]	format		data file format
		 * @return		true if success; false if failure
		 */
		bool readDataFile(const std::string &fileName, const std::string &format);

		/**
		 * Read similarity data file in the ASCII format.
		 * @param[in]	dataFile	data file stream
		 * @return		true if success; false if failure
		 */
		bool readDataFileAscii(std::ifstream &dataFile);

		/**
		 * Read similarity data file in Ali's format.
		 * @param[in]	dataFile	data file stream
		 * @return		true if success; false if failure
		 */
		bool readDataFileAli(std::ifstream &dataFile);

		/**
		 * Read class labels from file.
		 * @param[in]	fileName	label file name
		 * @return		true if success; false if failure
		 */
		bool readLabelFile(const std::string &fileName);

		/** Build the neighbor indices for all the samples. */
		void buildNeighborIdx();

		/** Build the map for class membership. */
		void buildClassMap();

		/**
		 * Get the indices of the training samples.
		 * @param[out]	trainingIdx		indices of the training samples
		 * @param[in]	testSampleIdx	index of the test sample (needed only when the data is
		 *								for leave-one-out cross-validation)
		 */
		void getTrainingSet(std::vector<unsigned int> &trainingIdx,
							unsigned int testSampleIdx = 0) const;

	private:
		unsigned int numSamples;	/**< Number of samples. */
		unsigned int numClasses;	/**< Number of classes. */

		bool normalizationFlag;		/**< Flag indicating if the data is normalized. */
		bool symmetryFlag;			/**< Flag indicating if the data is symmetric. */
		RealMatrix simMat;			/**< Similarity matrix. */
		std::vector<int> labels;	/**< Class labels of all the samples. */

		/** Matrix of neighbor indices for all the samples. */
		UIntMatrix neighborIdx;

		/** Type definition */
		typedef std::map< int, std::vector<unsigned int> > ClassMapType;
		ClassMapType classMap; /**< Map of class memebers. */

		bool loocvFlag; /**< Flag for leave-one-out cross-validation. */
		std::vector<unsigned int> trainingSet; /**< Training set indices. */
		std::vector<unsigned int> testSet; /**< Test set indices. */

		/** Matrix of neighbor indices for the test samples. */
		UIntMatrix testNeighborIdx;

		/** Matrix of neighbor indices for the test samples. */
		UIntMatrix trainingNeighborIdx;

	private:
		/** Structure for sorting based on the similarities. */
		struct _SortUnit
		{
			unsigned int idx;	/**< Sample index. */
			double val;			/**< Similarity value. */
		};

		/**
		 * Compare two samples based on their similarities.
		 * @param[in]	a	sample a
		 * @param[in]	b	sample b
		 * @return		true if a is more similar; false if b is more similar
		 */
		static bool _greater(_SortUnit a, _SortUnit b)
		{
			return a.val > b.val;
		}
	};
}

#endif // IDL_SIMDATA_H
