// File: SimTest.h
// Purpose: Definition of the SimTest class
// Remarks: This class is designed for classification test.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 1/5/2008 6:16:37 PM

#ifndef IDL_SIMTEST_H
#define IDL_SIMTEST_H

#include <string>
#include <vector>
#include <fstream>
#include "ConfigParser.h"
#include "SimData.h"
#include "ClassifierTest.h"

namespace idl
{
	/** Class for testing classifiers on similarity data. */
	class SimTest
	{
	public:
		/** Enumeration type for test. */
		enum TestType
		{
			TEST_RANDOM,		/**< test with random test set */
			TEST_FIXED,			/**< test with fixed test set */
			TEST_PRESPECIFIED	/**< test with prespecified test set */
		};

		/** Enumeration type for cross-validation. */
		enum CVType
		{
			CV_NOCV,			/**< no cross-validation */
			CV_LOOCV,			/**< leave-one-out cross-validation */
			CV_NFOLD,			/**< n-fold cross-validation with random partitions */
			CV_RANDOM,			/**< randomized cross-validation */
			CV_PRESPECIFIED		/**< cross-validation with prespecified partitions */
		};

	public:
		SimTest();	/**< Constructor. */
		~SimTest();	/**< Destructor. */

		void clear(); /**< Clear the current settings. */

		/**
		 * Start the test.
		 * @param[in]	config	ConfigParser object containing user configurations
		 * @return		true if success; false if failure
		 */
		bool start(ConfigParser &config);

	private:
		/**
		 * Configure the classification test.
		 * @param[in]	config	ConfigParser object containing user configurations
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config);

		/**
		 * Read the header of the partition file.
		 * @return	true if success; false if failure
		 */
		bool readPartitionFileHeader();

		/**
		 * Write the header of the partition file.
		 * @return	true if success; false if failure
		 */
		bool writePartitionFileHeader();

		/**
		 * Perform multiple rounds of tests.
		 * @param[in]	logFile		log file stream
		 * @param[in]	resultFile	result file stream
		 * @return		true if success; false if failure
		 */
		bool multipleTest(std::ofstream &logFile, std::ofstream &resultFile);

		/**
		 * Perform a single round of test.
		 * @param[in]	logFile		log file stream
		 * @param[in]	resultFile	result file stream
		 * @param[in]	testIdx		test index (0 by default)
		 * @return		true if success; false if failure
		 */
		bool singleTest(std::ofstream &logFile, std::ofstream &resultFile, unsigned int testIdx = 0);

		/**
		 * Perform cross-validation.
		 * @param[in]	logFile		log file stream
		 * @param[in]	resultFile	result file stream
		 * @param[in]	testIdx		test index (0 by default)
		 * @return		true if success; false if failure
		 */
		bool crossValidate(std::ofstream &logFile, std::ofstream &resultFile, unsigned int testIdx = 0);

		/**
		 * Load a test set from the partition file.
		 * @param[in]	testIdx	test index
		 * @return		true if success; false if failure
		 */
		bool loadTestSet(unsigned int testIdx);

		/**
		 * Save the test set to the partition file.
		 * @return	true if success; false if failure
		 */
		bool saveTestSet() const;

		/**
		 * Load cross-validation partitions from the partition file.
		 * @param[out]	testSets		indices of test samples for each partition
		 * @param[out]	trainingSets	indices of training samples for each partition
		 * @param[in]	testIdx			test index
		 * @return		true if success; false if failure
		 */
		bool loadPartitions(std::vector< std::vector<unsigned int> > &testSets,
							std::vector< std::vector<unsigned int> > &trainingSets,
							unsigned int testIdx) const;

		/**
		 * Save cross-validation partitions to the partition file.
		 * @param[in]	testSets	indices of test samples for each partition
		 * @return		true if success; false if failure
		 */
		bool savePartitions(const std::vector< std::vector<unsigned int> > &testSets) const;

		/**
		 * Save a cross-validation partition to the partition file.
		 * @param[in]	testIndices	indices of test samples for one partition
		 * @return		true if success; false if failure
		 */
		bool savePartition(const std::vector<unsigned int> &testIndices) const;

		/**
		 * Select a subset of the cross-validation partitions.
		 * @param[in, out]	testSets		indices of test samples for each partition
		 * @param[in, out]	trainingSets	indices of training samples for each partition
		 * @remarks			The subset is specified by vector cvSetIdx.
		 */
		void selectPartitions(std::vector< std::vector<unsigned int> > &testSets,
							  std::vector< std::vector<unsigned int> > &trainingSets) const;

		/**
		 * Log test information.
		 * @param[in]	file	output file stream
		 */
		void logTestInfo(std::ofstream &file) const;

		void getTrainingSet(); /**< Get the training set. */

		/**
		 * Get the complement of a subset.
		 * @param[in]	total		total number of samples
		 * @param[in]	subset		indices of the subset
		 * @param[out]	complement	indices of the complement of the given subset
		 * @attention	The indices of the subset must be in ascending order.
		 */
		static void getComplement(unsigned int total,
								  const std::vector<unsigned int> &subset,
								  std::vector<unsigned int> &complement);

		/**
		 * Get the relative complement of a set in another set.
		 * @param[in]	setA		indices of set A
		 * @param[in]	setB		indices of set B
		 * @param[out]	complement	indices of the relative complement of B in A, i.e., A\B
		 * @attention	The indices of both set A and set B must be in ascending order.
		 */
		static void getComplement(const std::vector<unsigned int> &setA,
								  const std::vector<unsigned int> &setB,
								  std::vector<unsigned int> &complement);

		/**
		 * Randomly divide the data into two parts.
		 * @param[in]	numSamples		total number of samples
		 * @param[in]	numPartition1	number of samples in partition 1
		 * @param[out]	partition1		indices of the samples in partition 1
		 * @param[out]	partition2		indices of the samples in partition 2
		 */
		static void randomPartition(unsigned int numSamples,
									unsigned int numPartition1,
									std::vector<unsigned int> &partition1,
									std::vector<unsigned int> &partition2);
		/**
		 * Randomly divide the data into two parts.
		 * @param[in]	samples			indices of all the samples
		 * @param[in]	numPartition1	number of samples in partition 1
		 * @param[out]	partition1		indices of the samples in partition 1
		 * @param[out]	partition2		indices of the samples in partition 2
		 */
		static void randomPartition(const std::vector<unsigned int> &samples,
									unsigned int numPartition1,
									std::vector<unsigned int> &partition1,
									std::vector<unsigned int> &partition2);

		/**
		 * Randomly divide the data into two parts.
		 * @param[in]	numSamples	total number of samples
		 * @param[in]	percentage	percentage of the samples in partition 1
		 * @param[out]	partition1	indices of the samples in partition 1
		 * @param[out]	partition2	indices of the samples in partition 2
		 */
		static void randomPartition(unsigned int numSamples,
									double percentage,
									std::vector<unsigned int> &partition1,
									std::vector<unsigned int> &partition2);

		/**
		 * Randomly divide the data into two parts.
		 * @param[in]	samples		indices of all the samples
		 * @param[in]	percentage	percentage of the samples in partition 1
		 * @param[out]	partition1	indices of the samples in partition 1
		 * @param[out]	partition2	indices of the samples in partition 2
		 */
		static void randomPartition(const std::vector<unsigned int> &samples,
									double percentage,
									std::vector<unsigned int> &partition1,
									std::vector<unsigned int> &partition2);

		/**
		 * Randomly divide the data into n equal parts.
		 * @param[in]	numSamples		total number of samples
		 * @param[in]	n				number of parts
		 * @param[out]	testSets		indices of test samples for each partition
		 * @param[out]	trainingSets	indices of training samples for each partition
		 */
		static void nfoldPartition(unsigned int numSamples,
								   unsigned int n,
								   std::vector< std::vector<unsigned int> > &testSets,
								   std::vector< std::vector<unsigned int> > &trainingSets);

		/**
		 * Randomly divide the data into n equal parts.
		 * @param[in]	samples			indices of all the samples
		 * @param[in]	n				number of parts
		 * @param[out]	testSets		indices of test samples for each partition
		 * @param[out]	trainingSets	indices of training samples for each partition
		 */
		static void nfoldPartition(const std::vector<unsigned int> &samples,
								   unsigned int n,
								   std::vector< std::vector<unsigned int> > &testSets,
								   std::vector< std::vector<unsigned int> > &trainingSets);

		/**
		 * Substitute the values for their indices.
		 * @param[in]	indices of all the samples
		 * @param[in]	multiple sets of indices
		 * @param[out]	multiple sets of values
		 */
		static void _substitute(const std::vector<unsigned int> &samples,
								const std::vector< std::vector<unsigned int> > &setsIdx,
								std::vector< std::vector<unsigned int> > &sets);

		/**
		 * Skip lines in the input file stream.
		 * @param[in]	file	input file stream
		 * @param[in]	nLines	number of lines to skip
		 */
		static void _skipLines(std::ifstream &file, unsigned int nLines);

	private:
		std::string dataFileName;		/**< Name of the data file. */
		std::string dataFormat;			/**< Format of the data file. */
		std::string labelFileName;		/**< Name of the label file. */
		std::string logFileName;		/**< Name of the log file. */
		std::string resultFileName;		/**< Name of the result file. */
		std::string partitionFileName;	/**< Name of the partition file. */

		/**
		 * Test level.
		 * Test level indicator: "1" for cross-validation only; "2" for cross-validation on
		 * the training set and test on the test set.
		 */
		int testLevel;

		TestType testMode; /**< Test mode. */

		/**
		 * Percentage of the test data.
		 * @attention	This is only for test with randomized training/test set.
		 */
		double testDataPercentage;

		/**
		 * Number of tests.
		 * @attention	This is only for test with randomized training/test set.
		 */
		unsigned int numTests;

		CVType cvMode; /**< Cross-validation mode. */

		/**
		 * Number of partitions for n-fold cross-validation.
		 * @attention	This is only for n-fold cross-validation.
		 */
		unsigned int cvNfold;

		/**
		 * Percentage of the test data used in cross-validation.
		 * @attention	This is only for randomized cross-validation.
		 */
		double cvPercentage;

		/**
		 * Number of tests in cross-validation.
		 * @attention	This is only for randomized cross-validation.
		 */
		unsigned int numCVs;

		/**
		 * Index of the test set in the partition file.
		 * @attention	This is only for prespecified test w/o cross-validation or
		 *				prespecified cross-validation.
		 */
		int testSetIdx;

		/**
		 * Indices of the cross-validation sets in the partition file.
		 * @attention	This is only for prespecified cross-validation.
		 */
		std::vector<unsigned int> cvSetIdx;

		SimData data;							/**< Similarity data. */
		std::vector<unsigned int> testSet;		/**< Indices of the test set. */
		std::vector<unsigned int> trainingSet;	/**< Indices of the training set. */

		/** Vector of pointers to the ClassifierTest objects. */
		std::vector<ClassifierTest *> classifierTests;
	};
}

#endif // IDL_SIMTEST_H
