// File: SimWeights.cpp
// Purpose: Implementation of the SimWeights class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/2/2008 11:57:54 AM

#include <iostream>
#include "LimeSolver.h"
#include "LimvSolver.h"
#include "SimWeights.h"

using namespace std;

namespace idl
{
	// Initialize the small positive number for regularization
	const double SimWeights::EPS = 1e-6;

	// Default constructor
	SimWeights::SimWeights()
		: nVariables(0), QMtx(NULL), qVec(NULL), wVec(NULL)
	{
	}

	// Constructor
	SimWeights::SimWeights(unsigned int n) : nVariables(n)
	{
		if (n > 1)
		{
			QMtx = new double[n * n];
			qVec = new double[n];
			wVec = new double[n];
		}
		else
		{
			QMtx = NULL;
			qVec = NULL;
			wVec = NULL;
		}
	}

	// Destructor
	SimWeights::~SimWeights()
	{
		delete [] QMtx;
		delete [] qVec;
		delete [] wVec;
	}

	// Compute the similarity weights
	void SimWeights::computeSimWeights(const RealMatrix	&simVec,
									   vector<double>	&weights)
	{
		unsigned int k = simVec.length();

		weights.resize(k);

		double s = simVec.sum();

		if (s <= 0.0)
		{
			for (unsigned int i = 0; i < k; ++i)
			{
				weights[i] = 1.0 / static_cast<double>(k);
			}
		}
		else
		{
			for (unsigned int i = 0; i < k; ++i)
			{
				weights[i] = simVec(i) / s;
			}
		}
	}

	// Compute the Markov weights
	// Attention: The similarities must be within the range of [0, 1].
	bool SimWeights::computeMarkovWeights(const RealMatrix	&simMat,
										  const RealMatrix	&simVec,
										  vector<double>	&weights,
										  double			lambda,
										  bool				dissim)
	{
		unsigned int k = simVec.length();

		RealMatrix P(k, k); // transition probability matrix

		// Form the transition probability matrix
		if (true == dissim)
		{
			// For the Markov weights that encourage diversity
			RealMatrix dissimMat(simMat);
			dissimMat.negate();	// dissimMat = -simMat
			dissimMat += 1.0;	// dissimMat = 1 - simMat

			for (unsigned int i = 0; i < k; ++i)
			{
				double r = 1.0 - simVec(i);
				double s = dissimMat.sumRow(i) - dissimMat(i, i);
				double t;

				if (s <= 0.0)
				{
					t = r / (static_cast<double>(k) - 1);
				}

				for (unsigned int j = 0; j < k; ++j)
				{
					if (j == i)
					{
						P(i, i) = simVec(i);
					}
					else
					{
						P(i, j) = (s <= 0.0) ? t : (dissimMat(i, j) / s) * r;
					}
				}
			}
		}
		else
		{
			// For the Markov weights that suppress diversity
			for (unsigned int i = 0; i < k; ++i)
			{
				double r = 1.0 - simVec(i);
				double s = simMat.sumRow(i) - simMat(i, i);
				double t;

				if (s <= 0.0)
				{
					t = r / (static_cast<double>(k) - 1);
				}

				for (unsigned int j = 0; j < k; ++j)
				{
					if (j == i)
					{
						P(i, i) = simVec(i);
					}
					else
					{
						P(i, j) = (s <= 0.0) ? t : (simMat(i, j) / s) * r;
					}
				}
			}
		}

		// regularization: P = (1 - lambda) * P + lambda * ones(k) / k
		P *= (1 - lambda);
		P += lambda / static_cast<double>(k);

		RealMatrix A(k, k);
		A.eye();
		A -= P.transpose();  // A = I - P'
		A.fillRow(k - 1, 1.0);

		RealMatrix b(k, 1, 0.0);
		b(k - 1) = 1.0;

		RealMatrix x;
		if (!solve(A, b, x)) // Solve for the stationary probabilities
		{
			cerr << "Error: Cannot solve the linear system for the Markov weights" << endl;
			return false;
		}
		x.dump(weights);

		return true;
	}

	// Resize the buffer
	void SimWeights::resizeBuffer(unsigned int k)
	{
		if (nVariables != k)
		{
			nVariables = k;

			delete [] QMtx;
			delete [] qVec;
			delete [] wVec;

			if (k > 0)
			{
				QMtx = new double[k * k];
				qVec = new double[k];
				wVec = new double[k];
			}
			else
			{
				QMtx = NULL;
				qVec = NULL;
				wVec = NULL;
			}
		} 
	}

	// Generate matrix Q
	void SimWeights::generateQMtx(const RealMatrix &simMat, ApproxType approx, bool symFlag)
	{
		unsigned int k = simMat.getRowNum();

		RealMatrix Q;
		if (symFlag)
		{
			Q = simMat;
		}
		else
		{
			Q = simMat + simMat.transpose();
			Q *= 0.5; // Q = (Sx + Sx') / 2
		}

		RealMatrix eigVecs;			// V'
		RealMatrix eigVals;			// D
		Q.eig(eigVecs, eigVals);	// Eigendecomposition: Q = VDV'

		double minEigVal = eigVals.min();

		if (minEigVal < 0.0)
		{
			if (APPROX_EIG_SHIFT == approx)
			{
				// Add a small regularization term to counter numerical instability
				double delta = EPS - minEigVal;

				// Make Q positive semidefinite by shifting the eigenvalues
				for (unsigned int i = 0; i < k; ++i)
				{
					Q(i, i) += delta;
				}
			}
			else
			{
				RealMatrix V = eigVecs.transpose();

				if (APPROX_EIG_CLIP == approx)
				{
					// Make Q positive semidefinite by clipping the negative eigenvalues
					for (unsigned int i = 0; i < k; ++i)
					{
						if (eigVals(i) < 0.0) eigVals(i) = EPS;
					}
				}
				else // APPROX_EIG_FLIP == approx
				{
					// Make Q positive semidefinite by flipping the sign of the negative eigenvalues
					for (unsigned int i = 0; i < k; ++i)
					{
						if (eigVals(i) < 0.0) eigVals(i) = EPS - eigVals(i);
					}
				}

				for (unsigned int i = 0; i < k; ++i)
				{
					for (unsigned int j = 0; j < k; ++j)
					{
						eigVecs(i, j) *= eigVals(i);
					}
				}

				Q = V * eigVecs;
			}
		}

		// Copy the elements to the buffer
		for (unsigned int i = 0; i < k * k; ++i)
		{
			QMtx[i] = Q(i);
		}
	}

	// Compute the sim-LIME weights
	bool SimWeights::computeLimeWeights(const RealMatrix	&simMat,
										const RealMatrix	&simVec,
										vector<double>		&weights,
										double				lambda,
										ApproxType			approx,
										bool				symFlag)
	{
		unsigned int k = simMat.getRowNum();
		resizeBuffer(k);

		generateQMtx(simMat, approx, symFlag);

		// Generate vector q
		for (unsigned int i = 0; i < k; ++i)
		{
			qVec[i] = -simVec(i); // q = -Sz'
		}

		// Solve the LIME problem
		if (!LimeSolver::solve(static_cast<int>(k), QMtx, qVec, lambda, wVec)) return false;

		weights.assign(wVec, wVec + k);

		return true;
	}

	// Compute the sim-LIMV weights
	bool SimWeights::computeLimvWeights(const RealMatrix	&simMat,
										const RealMatrix	&simVec,
										vector<double>		&weights,
										double				lambda,
										ApproxType			approx,
										bool				symFlag)
	{
		unsigned int k = simMat.getRowNum();
		resizeBuffer(k);

		generateQMtx(simMat, approx, symFlag);

		// Generate vector q
		for (unsigned int i = 0; i < k; ++i)
		{
			qVec[i] = -simVec(i); // q = -Sz'
		}

		// Solve the LIMV problem
		if (!LimvSolver::solve(static_cast<int>(k), QMtx, qVec, lambda, wVec)) return false;

		weights.assign(wVec, wVec + k);

		return true;
	}

	// Compute the linear regression weights
	bool SimWeights::computeLinearRegressionWeights(const RealMatrix	&simMat,
													const RealMatrix	&simVec,
													vector<double>		&weights,
													double				lambda,
													ApproxType			approx,
													bool				symFlag)
	{
		unsigned int k = simMat.getRowNum();
		resizeBuffer(k);

		RealMatrix A(k, k);	// coefficient matrix
		RealMatrix b = simVec.transpose();

		if (APPROX_NONE != approx)
		{
			// spectrum clip, flip, & shift
			generateQMtx(simMat, approx, symFlag);
			for (unsigned int i = 0; i < k * k; ++i)
			{
				A(i) = QMtx[i];
			}
		}
		else // pseudoinverse
		{
			if (symFlag)
			{
				A = simMat;
			}
			else
			{
				A = simMat + simMat.transpose();
				A *= 0.5; // A = (Sx + Sx') / 2
			}
		}

		for (unsigned int i = 0; i < k; ++i)
		{
			A(i, i) += lambda;
		}

		RealMatrix x;
		if (!solve(A, b, x, true)) // Solve the linear system: Ax = b
		{
			cerr << "Error: Cannot solve the linear system for the linear regression weights" << endl;
			return false;
		}
		x += (1.0 / static_cast<double>(k) - x.mean()); // Normalize the weights
		x.dump(weights);

		return true;
	}
}
