// File: WeightedKNN.cpp
// Purpose: Implementation of the WeightedKNN class
// Remarks: This class is designed for weighted k-NN classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/2/2008 12:04:05 PM

//#include <cstdlib>
//#include <ctime>
#include <cfloat>
#include <map>
#include <iostream>
#include "CmRandomNumberGenerator.h" // Mersenne Twister (singleton) header file
#include "RealMatrix.h"
#include "WeightedKNN.h"
#include <assert.h>

using namespace std;

namespace idl
{
	// Initialize the string representations of the weighting schemes
	const char *WeightedKNN::weightNames[] = {
		"uniform", "similarity", "Markov (sim)", "Markov (dissim)",
		"sim-LIME (shift)", "sim-LIME (clip)", "sim-LIME (flip)",
		"sim-LIMV (shift)", "sim-LIMV (clip)", "sim-LIMV (flip)",
		"linear regression (pinv)", "linear regression (shift)",
		"linear regression (clip)", "linear regression (flip)" };

	// Set the weight mode
	bool WeightedKNN::setWeightMode(string &strWeightMode)
	{
		if ("uniform" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_UNIFORM;
		}
		else if ("sim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_SIM;
		}
		else if ("markov_sim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_MARKOV_SIM;
		}
		else if ("markov_dissim" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_MARKOV_DISSIM;
		}
		else if ("lime_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_SHIFT;
		}
		else if ("lime_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_CLIP;
		}
		else if ("lime_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIME_FLIP;
		}
		else if ("limv_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_SHIFT;
		}
		else if ("limv_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_CLIP;
		}
		else if ("limv_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LIMV_FLIP;
		}
		else if ("lr_pinv" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_PINV;
		}
		else if ("lr_shift" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_SHIFT;
		}
		else if ("lr_clip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_CLIP;
		}
		else if ("lr_flip" == strWeightMode)
		{
			parameters.weightMode = WEIGHT_LR_FLIP;
		}
		else
		{
			return false;
		}

		return true;
	}

	// Return the probability of each class with non-uniform weights.
	void WeightedKNN::getClassProbs(const vector<int> &labels, const vector<double> &weights, const vector<int> &uniqueClassLabels, map<int,double> &probs)
	{
		vector<int>::const_iterator itrL;
		vector<double>::const_iterator itrW;
		map<int, double> counts;
		counts.clear();

		// Count weighted class occurrences among the neighbors
		for (itrL = labels.begin(), itrW = weights.begin();
			itrL != labels.end() && itrW != weights.end();
			++itrL, ++itrW)
		{
			if (counts.find(*itrL) == counts.end())
			{
				counts[*itrL] = *itrW;
			}
			else
			{
				counts[*itrL] += *itrW;
			}
		}
		// sum the counts
		double sumCounts = 0;
		for (map<int, double>::const_iterator itr = counts.begin();
			itr != counts.end(); ++itr)
		{
			sumCounts += itr->second;
		}
		// probs[i] = counts[i]/sum(counts)
		for (vector<int>::const_iterator itrL = uniqueClassLabels.begin();
			 itrL != uniqueClassLabels.end(); ++itrL)
		{
			probs[*itrL] = counts[*itrL]/sumCounts;
		}
	}

	// Return the counts of each label with uniform weights
	void WeightedKNN::getClassProbs(const vector<int> &labels, const vector<int> &uniqueClassLabels, map<int,double> &probs)
	{
		const vector<double> weights(labels.size(), 1);
		
		getClassProbs(labels, weights, uniqueClassLabels, probs);
	}


	// Classify the test samples of the similarity data set
	bool WeightedKNN::classify(const SimData &data, vector<int> &results, RealMatrix &probMatrix)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		if (data.getTrainingSetSize() < parameters.neighborhoodSize)
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples, number of unique labels
		unsigned int numTestSamples = data.getTestSetSize();
		vector<int> uniqueClassLabels;
		data.getUniqueLabels(uniqueClassLabels);
		probMatrix.resize(numTestSamples,(unsigned int)uniqueClassLabels.size());
		results.resize(numTestSamples);

		if (WEIGHT_UNIFORM == parameters.weightMode)
		{
			// Uniform weights
			vector<int> labels(parameters.neighborhoodSize);
			map<int, double> probs;

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNLabels(i, parameters.neighborhoodSize, labels);
				// Get class probabilities for this sample.
				getClassProbs(labels, uniqueClassLabels, probs);
				// Save probabilities to matrix.
				unsigned int idx_label = 0;
				for ( vector<int>::const_iterator itrL = uniqueClassLabels.begin();
					itrL != uniqueClassLabels.end(); ++itrL)
				{
					probMatrix(i,idx_label) = probs[*itrL];
					++idx_label;
				}
				results[i] = vote(labels);
			}
		}
		else if (WEIGHT_SIM == parameters.weightMode)
		{
			// Similarity weights
			RealMatrix simVec(1, parameters.neighborhoodSize);
			vector<int> labels(parameters.neighborhoodSize);
			vector<double> weights(parameters.neighborhoodSize);
			map<int,double> probs;

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
				data.getNNLabels(i, parameters.neighborhoodSize, labels);

				simWeights.computeSimWeights(simVec, weights);

				// Get class probabilities for this sample.
				getClassProbs(labels, weights, uniqueClassLabels, probs);
				// Save probabilities to matrix.
				unsigned int idx_label = 0;
				for ( vector<int>::const_iterator itrL = uniqueClassLabels.begin();
					itrL != uniqueClassLabels.end(); ++itrL)
				{
					probMatrix(i,idx_label) = probs[*itrL];
					++idx_label;
				}

				results[i] = vote(labels, weights);
			}
		}
		else // Markov, LIME, LIMV and linear regression weights
		{
			RealMatrix simMat(parameters.neighborhoodSize,
							  parameters.neighborhoodSize);
			RealMatrix simVec(1, parameters.neighborhoodSize);
			vector<int> labels(parameters.neighborhoodSize);
			vector<double> weights(parameters.neighborhoodSize);
			map<int,double> probs;

			bool symFlag = data.isSymmetric();

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNSimMat(i, parameters.neighborhoodSize, simMat);
				data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
				data.getNNLabels(i, parameters.neighborhoodSize, labels);

				if (!data.isNormalized())
				{
					// Normalize the similarities locally
					double m1 = simMat.max();
					double m2 = simVec.max();
					double m3 = (m1 > m2) ? m1 : m2;

					simMat /= m3;
					simVec /= m3;
				}

				switch(parameters.weightMode)
				{
				case WEIGHT_MARKOV_DISSIM:
					// Markov weights that encourage diversity
					if (!simWeights.computeMarkovWeights(simMat, simVec, weights,
						parameters.lambda, true)) return false;
					break;

				case WEIGHT_MARKOV_SIM:
					// Markov weights that suppress diversity
					if (!simWeights.computeMarkovWeights(simMat, simVec, weights,
						parameters.lambda, false)) return false;
					break;

				case WEIGHT_LIME_SHIFT:
					// sim-LIME weights with eigenvalue shift
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LIME_CLIP:
					// sim-LIME weights with eigenvalue clip
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIME_FLIP:
					// sim-LIME weights with eigenvalue flip
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_SHIFT:
					// sim-LIMV weights with eigenvalue shift
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_CLIP:
					// sim-LIMV weights with eigenvalue clip
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_FLIP:
					// sim-LIMV weights with eigenvalue flip
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;

				case WEIGHT_LR_PINV:
					// linear regression weights with pseudoinverse
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_NONE, symFlag))
						return false;
					break;

				case WEIGHT_LR_SHIFT:
					// linear regression weights with eigenvalue shift
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LR_CLIP:
					// linear regression weights with eigenvalue clip
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LR_FLIP:
					// linear regression weights with eigenvalue flip
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;
				}

				// Get class probabilities for this sample.
				getClassProbs(labels, weights, uniqueClassLabels, probs);
				// Save probabilities to matrix.
				unsigned int idx_label = 0;
				for ( vector<int>::const_iterator itrL = uniqueClassLabels.begin();
					itrL != uniqueClassLabels.end(); ++itrL)
				{
					probMatrix(i,idx_label) = probs[*itrL];
					++idx_label;
				}

				results[i] = vote(labels, weights);
			}
		}

		double temp = probMatrix.sum();
		assert(probMatrix.sum() < numTestSamples + .000001);
		assert(probMatrix.sum() > numTestSamples - .000001);
		return true;
	}

	// Classify the test samples of the similarity data set
	bool WeightedKNN::classify(const SimData &data, vector<int> &results)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		if (data.getTrainingSetSize() < parameters.neighborhoodSize)
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		if (1 == parameters.neighborhoodSize)
		{
			// 1-NN
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				results[i] = data.get1NNLabel(i);
			}
		}
		else if (WEIGHT_UNIFORM == parameters.weightMode)
		{
			// Uniform weights
			vector<int> labels(parameters.neighborhoodSize);

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNLabels(i, parameters.neighborhoodSize, labels);
				results[i] = vote(labels);
			}
		}
		else if (WEIGHT_SIM == parameters.weightMode)
		{
			// Similarity weights
			RealMatrix simVec(1, parameters.neighborhoodSize);
			vector<int> labels(parameters.neighborhoodSize);
			vector<double> weights(parameters.neighborhoodSize);

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
				data.getNNLabels(i, parameters.neighborhoodSize, labels);

				simWeights.computeSimWeights(simVec, weights);

				results[i] = vote(labels, weights);
			}
		}
		else // Markov, LIME, LIMV and linear regression weights
		{
			RealMatrix simMat(parameters.neighborhoodSize,
							  parameters.neighborhoodSize);
			RealMatrix simVec(1, parameters.neighborhoodSize);
			vector<int> labels(parameters.neighborhoodSize);
			vector<double> weights(parameters.neighborhoodSize);

			bool symFlag = data.isSymmetric();

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNSimMat(i, parameters.neighborhoodSize, simMat);
				data.getNNSimVec(i, parameters.neighborhoodSize, simVec);
				data.getNNLabels(i, parameters.neighborhoodSize, labels);

				if (!data.isNormalized())
				{
					// Normalize the similarities locally
					double m1 = simMat.max();
					double m2 = simVec.max();
					double m3 = (m1 > m2) ? m1 : m2;

					simMat /= m3;
					simVec /= m3;
				}

				switch(parameters.weightMode)
				{
				case WEIGHT_MARKOV_DISSIM:
					// Markov weights that encourage diversity
					if (!simWeights.computeMarkovWeights(simMat, simVec, weights,
						parameters.lambda, true)) return false;
					break;

				case WEIGHT_MARKOV_SIM:
					// Markov weights that suppress diversity
					if (!simWeights.computeMarkovWeights(simMat, simVec, weights,
						parameters.lambda, false)) return false;
					break;

				case WEIGHT_LIME_SHIFT:
					// sim-LIME weights with eigenvalue shift
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LIME_CLIP:
					// sim-LIME weights with eigenvalue clip
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIME_FLIP:
					// sim-LIME weights with eigenvalue flip
					if (!simWeights.computeLimeWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_SHIFT:
					// sim-LIMV weights with eigenvalue shift
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_CLIP:
					// sim-LIMV weights with eigenvalue clip
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LIMV_FLIP:
					// sim-LIMV weights with eigenvalue flip
					if (!simWeights.computeLimvWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;

				case WEIGHT_LR_PINV:
					// linear regression weights with pseudoinverse
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_NONE, symFlag))
						return false;
					break;

				case WEIGHT_LR_SHIFT:
					// linear regression weights with eigenvalue shift
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_SHIFT, symFlag))
						return false;
					break;

				case WEIGHT_LR_CLIP:
					// linear regression weights with eigenvalue clip
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_CLIP, symFlag))
						return false;
					break;

				case WEIGHT_LR_FLIP:
					// linear regression weights with eigenvalue flip
					if (!simWeights.computeLinearRegressionWeights(simMat, simVec, weights,
						parameters.lambda, SimWeights::APPROX_EIG_FLIP, symFlag))
						return false;
					break;
				}

				results[i] = vote(labels, weights);
			}
		}

		return true;
	}

	// Vote with uniform weights
	int WeightedKNN::vote(const vector<int> &labels)
	{
		map<int, unsigned int> counts;

		// Count class occurrences among the neighbors
		for (vector<int>::const_iterator itr = labels.begin();
			 itr != labels.end(); ++itr)
		{
			if (counts.find(*itr) == counts.end())
			{
				counts[*itr] = 1;
			}
			else
			{
				++(counts[*itr]);
			}
		}

		unsigned int maxCount = 0;

		// Find the maximum occurrence
		for (map<int, unsigned int>::const_iterator itr = counts.begin();
			 itr != counts.end(); ++itr)
		{
			if (itr->second > maxCount)
			{
				maxCount = itr->second;
			}
		}

		vector<int> candidates;

		// Find the classes with the maximum occurrence
		for (map<int, unsigned int>::const_iterator itr = counts.begin();
			 itr != counts.end(); ++itr)
		{
			if (itr->second == maxCount)
			{
				candidates.push_back(itr->first);
			}
		}

		int winner;

		// Make the final decision
		if (1 == candidates.size())
		{
			// Only one class with the maximum occurrence
			winner = candidates[0];
		}
		else
		{
			// Multiple classes with the maximum occurrence:
			// Choose one randomly
			//srand(static_cast<unsigned int>(time(NULL)));
			//winner = candidates[rand() % candidates.size()];
			haruki::CmRandomNumberGenerator *pRandomNumberGenerator =
				haruki::CmRandomNumberGenerator::getInstance();
			winner = candidates[pRandomNumberGenerator->getUInt() % candidates.size()];
		}

		return winner;
	}

	// Vote with non-uniform weights
	int WeightedKNN::vote(const vector<int> &labels, const vector<double> &weights)
	{
		map<int, double> counts;

		vector<int>::const_iterator itrL;
		vector<double>::const_iterator itrW;

		// Count weighted class occurrences among the neighbors
		for (itrL = labels.begin(), itrW = weights.begin();
			 itrL != labels.end() && itrW != weights.end();
			 ++itrL, ++itrW)
		{
			if (counts.find(*itrL) == counts.end())
			{
				counts[*itrL] = *itrW;
			}
			else
			{
				counts[*itrL] += *itrW;
			}
		}

		double maxCount = -DBL_MAX;

		// Find the maximum weighted occurrence
		for (map<int, double>::const_iterator itr = counts.begin();
			 itr != counts.end(); ++itr)
		{
			if (itr->second > maxCount)
			{
				maxCount = itr->second;
			}
		}

		vector<int> candidates;

		// Find the classes with the maximum weighted occurrence
		for (map<int, double>::const_iterator itr = counts.begin();
			 itr != counts.end(); ++itr)
		{
			if (itr->second == maxCount)
			{
				candidates.push_back(itr->first);
			}
		}

		int winner;

		// Make the final decision
		if (1 == candidates.size())
		{
			// Only one class with the maximum weighted occurrence
			winner = candidates[0];
		}
		else
		{
			// Multiple classes with the maximum weighted occurrence:
			// Choose one randomly
			//srand(static_cast<unsigned int>(time(NULL)));
			//winner = candidates[rand() % candidates.size()];
			haruki::CmRandomNumberGenerator *pRandomNumberGenerator =
				haruki::CmRandomNumberGenerator::getInstance();
			winner = candidates[pRandomNumberGenerator->getUInt() % candidates.size()];
		}

		return winner;
	}

	
}
