// File: WeightedKNNBN.h
// Purpose: Definition of the WeightedKNNBN class
// Remarks: This class is designed for weighted k-NN classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/24/2008 4:23:48 PM

#ifndef IDL_WEIGHTEDKNNBN_H
#define IDL_WEIGHTEDKNNBN_H

#include <string>
#include <vector>
#include "Classifier.h"
#include "SimData.h"
#include "SimWeights.h"
#include "WeightedKNN.h"

namespace idl
{
	/**
	 * Class for the weighted k-NN classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class WeightedKNNBN : public WeightedKNN
	{
	public:
		///* Enumeration type for weighting scheme. */
		////enum WeightType
		//{
		//	WEIGHT_UNIFORM,			/**< uniform weights */
		//	WEIGHT_SIM,				/**< similarity weights */
		//	WEIGHT_MARKOV_SIM,		/**< Markov weights suppressing diversity */
		//	WEIGHT_MARKOV_DISSIM,	/**< Markov weights encouraging diversity */
		//	WEIGHT_LIME_SHIFT,		/**< sim-LIME weights with eigenvalue shift */
		//	WEIGHT_LIME_CLIP,		/**< sim-LIME weights with eigenvalue clip */
		//	WEIGHT_LIME_FLIP,		/**< sim-LIME weights with eigenvalue flip */
		//	WEIGHT_LIMV_SHIFT,		/**< sim-LIMV weights with eigenvalue shift */
		//	WEIGHT_LIMV_CLIP,		/**< sim-LIMV weights with eigenvalue clip */
		//	WEIGHT_LIMV_FLIP,		/**< sim-LIMV weights with eigenvalue flip */
		//	WEIGHT_LR_PINV,			/**< linear weights with pseudoinverse */
		//	WEIGHT_LR_SHIFT,		/**< linear weights with eigenvalue shift */
		//	WEIGHT_LR_CLIP,			/**< linear weights with eigenvalue clip */
		//	WEIGHT_LR_FLIP			/**< linear weights with eigenvalue flip */
		//};

		/** Structure type for the weighted k-NN parameters. */
		struct ParaType
		{
			//unsigned int neighborhoodSize;	/**< Neighborhood size. */
			std::vector<unsigned int> neighborhoodSizes;
			WeightType weightMode;			/**< Weight mode. */
			double lambda;					/**< Regularization parameter. */
		};

		/** Default constructor. */
		WeightedKNNBN()
		{
			//parameters.neighborhoodSize = 1;		// default neighborhood size: 1
			//vector<unsigned int> neighborhoodSizes; 
			parameters.weightMode = WEIGHT_UNIFORM;	// default weighting scheme: UNIFORM
			parameters.lambda = 0.0;				// default regularization parameter: 0.0
		}

		/**
		 * Constructor.
		 * @param	para parameters for the weighted k-NN classifier
		 */
		WeightedKNNBN(const ParaType &para)
		{
			parameters = para;
		}

		~WeightedKNNBN() {} /**< Destructor. */

		/**
		 * Set the parameters for the weighted k-NN classifier.
		 * @param[in]	para	parameters for the weighted k-NN classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSizes(std::vector<unsigned int> &nbSizes)
		{
			//if (k > 0)
			//{
			parameters.neighborhoodSizes = nbSizes;
			//}
		}

		/**
		 * Set the regularization parameter.
		 * @param[in]	p	regularization parameter
		 */
		void setRegPara(double p)
		{
			parameters.lambda = p;
		}

		/**
		 * Set the weight mode.
		 * @param[in]	t	weight mode
		 */
		void setWeightMode(WeightType t)
		{
			parameters.weightMode = t;
		}

		/**
		 * Set the weight mode.
		 * @param[in]	strWeightMode	string representing the weight mode
		 * @return		true if success; false if failure
		 */
		bool setWeightMode(std::string &strWeightMode);

		/**
		 * Get the name of the weight mode.
		 * @return	string representation of the weight mode
		 */
		const char *getWeightName() const
		{
			return weightNames[parameters.weightMode];
		}

		/**
		 * Check if the regularization parameter is needed.
		 * @return	true if needed; false if not
		 */
		bool needRegPara() const
		{
			return parameters.weightMode > WEIGHT_SIM;
		}

		/**
		 * Check the legitimacy of the range of the regularization parameters.
		 * @param[in]	lowerBound		lower bound of the range
		 * @param[in]	upperBound		upper bound of the range
		 * @return		true if the range is legitimate; false if not
		 * @attention	The input arguments must satisfy lowerBound <= upperBound.
		 */
		bool checkRegPara(double lowerBound, double upperBound) const
		{
			if (WEIGHT_MARKOV_SIM == parameters.weightMode
				|| WEIGHT_MARKOV_DISSIM == parameters.weightMode)
			{
				// Legal range for Markov weights: (0, 1]
				return lowerBound > 0.0 && upperBound <= 1.0;
			}
			else if (parameters.weightMode > WEIGHT_MARKOV_DISSIM)
			{
				// Legal range for sim-LIME/sim-LIMV weights: (0, +\infty)
				return lowerBound > 0.0;
			}

			return true;
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data	similarity data set
		 * @param[out]	results	vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);
		bool classify(const SimData &data, std::vector<int> &results, RealMatrix &probMatrix);

	private:
		/**
		 * Vote with uniform weights.
		 * @param[in]	labels	class labels of the neighbors
		 * @return		estimated class label of the test sample
		 */
		//static int vote(const std::vector<int> &labels);

		/**
		 * Vote with non-uniform weights.
		 * @param[in]	labels	class labels of the neighbors
		 * @param[in]	weights	weights of the neighbors
		 * @return		estimated class label of the test sample
		 */
		//static int vote(const std::vector<int> &labels, const std::vector<double> &weights);

		//bool getProbabilities(const SimData &data, unsigned int testIdx, RealMatrix &probabilities);
		//void getCounts(const std::vector<int> &labels, std::map<int,double> &counts);
		//void getCounts(const std::vector<int> &labels, const std::vector<double> &weights, std::map<int,double> &counts);

		int maxLabel(const std::vector<int> &labels, const RealMatrix &probabilities);

		/** String representations of the weighting schemes. */
		static const char *weightNames[];

	private:
		ParaType parameters;	/**< Parameters for the weighted k-NN classifier. */
		//SimWeights simWeights;	/**< Object for computing the weights of the neighbors. */
	};
}

#endif // IDL_WEIGHTEDKNN_H
