// File: WeightedKNNTest.cpp
// Purpose: Implementation of the WeightedKNNTest class
// Remarks: This class is designed as an adapter for the WeightedKNN class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:24:41 PM

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "WeightedKNNTest.h"

using namespace std;

namespace idl
{
	const string WeightedKNNTest::_ID = "kNN"; // Define the class ID

	// Register the class instantiator with the factory
	const bool WeightedKNNTest::_REGISTERED =
		ClassifierTestFactory::registerCreator(WeightedKNNTest::_ID, &WeightedKNNTest::create);

	// Destructor
	WeightedKNNTest::~WeightedKNNTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a weighted k-NN classifier
	bool WeightedKNNTest::configure(ConfigParser &config, int secIdx)
	{
		string strWeightMode;

		// Get weight mode for weighted k-NN
		if (!config.getString(secIdx, "WeightMode", strWeightMode))
		{
			cerr << "Missing weight mode for k-NN classifier " << secIdx << endl;
			return false;
		}
		if (!knn.setWeightMode(strWeightMode))
		{
			cerr << "Unknown weight mode for k-NN classifier " << secIdx << endl;
			return false;
		}
		
		if (knn.needRegPara())
		{
			// Get regularization parameters to cross-validate
			if (!config.getRealVec(secIdx, "lambda", lambdas))
			{
				cerr << "Error: Missing or incorrect lambda for k-NN classifier "
					 << secIdx << endl;
				return false;
			}

			// Check the retrieved regularization parameters
			if (!checkLambdas())
			{
				cerr << "Error: Incorrect lambda for k-NN classifier "
					 << secIdx << endl;
				return false;
			}
		}
		else
		{
			lambdas.clear();
			lambdas.push_back(0.0);
		}

		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for kNN classifier "
				 << secIdx << endl;
			return false;
		}

		// Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for kNN classifier "
				 << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(lambdas.size() * neighborhoodSizes.size()));

		return true;
	}

	// Test the weighted k-NN classifier on all available parameters
	bool WeightedKNNTest::test(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		logFile << "Weighted k-NN classifier" << endl
				<< "Weight mode: " << knn.getWeightName() << endl
				<< "Neighborhood size\tLambda\tError rate" << endl;

		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());

		for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
			 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
		{
			knn.setNeighborhoodSize(*itrNeighborhoodSize);

			for (vector<double>::const_iterator itrLambda = lambdas.begin();
				 itrLambda != lambdas.end(); ++itrLambda)
			{
				knn.setRegPara(*itrLambda);
				if (!knn.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);

				// Log the error rate
				logFile << *itrNeighborhoodSize << "\t" << *itrLambda << "\t"
						<< (errorRate * 100.0) << "%" << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			}
		}

		logFile << endl;

		++testCount;

		return true;
	}

	// Check the legitimacy of the regularization parameters
	bool WeightedKNNTest::checkLambdas()
	{
		sort(lambdas.begin(), lambdas.end());

		return knn.checkRegPara(lambdas.front(), lambdas.back());
	}

	// Check the legitimacy of the neighborhood sizes
	bool WeightedKNNTest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void WeightedKNNTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void WeightedKNNTest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.lambda = lambdas.front();
			latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize = neighborhoodSizes[i / lambdas.size()];

			// Get the lambda corresponding to the minimum cross-validation error
			latestResult.lambda = lambdas[i % lambdas.size()];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the weighted k-NN classifier with the best parameters
	bool WeightedKNNTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		logFile << "Weighted k-NN classifier" << endl
				<< "Weight mode: " << knn.getWeightName() << endl;

		unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		knn.setNeighborhoodSize(k);

		double lambda = (resultsBestPara.back()).lambda;
		knn.setRegPara(lambda);

		vector<int> results(data.getTestSetSize());
		RealMatrix probMatrix;
		if (!knn.classify(data, results, probMatrix)) return false;

		// Log the probabilities.
		logFile << "Classification probabilities:" << endl;
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				logFile << probMatrix(i,j) << " ";
			}
			logFile << endl;
		}
		logFile << "Name to help with saving:" << endl
				<< knn.getWeightName() << "k" << k << "lambda" << lambda << endl
				<< endl;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Neighborhood size\tLambda\tError rate" << endl
				<< k << "\t" << lambda << "\t" << (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void WeightedKNNTest::reportResults(ofstream &file, bool verbose)
	{
		file << "Weighted k-NN classifier" << endl
			 << "Weight mode: " << knn.getWeightName() << endl;

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "Neighborhood size\tLambda\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				for (vector<double>::const_iterator itrLambda = lambdas.begin();
					 itrLambda != lambdas.end(); ++itrLambda)
				{
					file << *itrNeighborhoodSize << "\t" << *itrLambda << "\t"
						 << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
					++i;
				}
			}
		}
		else
		{
			// Report the best result only
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Lambda: " << (resultsBestPara.back()).lambda << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void WeightedKNNTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "Weighted k-NN classifier" << endl
			 << "Weight mode: " << knn.getWeightName() << endl;

		if (verbose)
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Lambda: " << (resultsBestPara.back()).lambda << endl
				 << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Lambda: " << (resultsBestPara.back()).lambda << endl
				 << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void WeightedKNNTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "Weighted k-NN classifier" << endl
			 << "Weight mode: " << knn.getWeightName() << endl;

		if (verbose)
		{
			file << "Neighborhood size\tLambda\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix lambdaVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			lambdaVec(i) = itr->lambda;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->neighborhoodSize << "\t" << itr->lambda << "\t"
					 << (itr->testError * 100.0) << "%" << endl;
			}
		}
		
		file << "Average neighborhood size: " << kVec.mean()
			 << " (" << kVec.std() << ")" << endl
			 << "Average lambda: " << lambdaVec.mean()
			 << " (" << lambdaVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
